"""Tool management for the Arklex framework.

This module provides functionality for managing tools, including
initialization, execution, and slot filling integration.
"""

import asyncio
import inspect
import json
import traceback
import uuid
from collections.abc import Callable
from typing import Any

from agents import FunctionTool, RunContextWrapper
from pydantic import BaseModel, Field, create_model

from arklex.orchestrator.entities.orchestrator_state_entities import (
    OrchestratorState,
    StatusEnum,
)
from arklex.orchestrator.NLU.core.slot import SlotFiller
from arklex.orchestrator.NLU.entities.slot_entities import (
    Slot,
)
from arklex.utils.exceptions import AuthenticationError, ToolExecutionError
from arklex.utils.logging_utils import LogContext
from arklex.utils.utils import format_chat_history

log_context = LogContext(__name__)


class ToolOutput(BaseModel):
    status: StatusEnum
    message_flow: str | None = None
    response: str | None = None
    slots: dict[str, list[Slot]] | None = None


# Type conversion mapping for slot values
TYPE_CONVERTERS = {
    "int": int,
    "float": float,
    "bool": lambda v: v
    if isinstance(v, bool)
    else (v.lower() == "true" if isinstance(v, str) else bool(v)),
    "str": lambda v: v if isinstance(v, dict | list) else str(v),
}


def register_tool(
    description: str,
    slots: list[dict[str, Any]] | None = None,
) -> Callable:
    """Register a tool with the Arklex framework.

    This decorator registers a function as a tool with the specified description, slots,
    outputs, and response flag. It handles path normalization and tool initialization.

    Args:
        desc (str): Description of the tool's functionality.
        slots (List[Dict[str, Any]], optional): List of slot definitions. Defaults to None.

    Returns:
        Callable: A function that creates and returns a Tool instance.
    """
    if slots is None:
        slots = []

    def inner(func: Callable) -> Callable:
        name: str = f"{func.__name__}"
        return Tool(func, name, description, slots)

    return inner


class Tool:
    """Base class for tools in the Arklex framework.

    This class provides the core functionality for tool execution, slot management,
    and state handling. It supports slot filling, parameter validation, and error
    handling during tool execution.

    Attributes:
        func (Callable): The function implementing the tool's functionality.
        name (str): The name of the tool.
        description (str): Description of the tool's functionality.
        output (List[str]): List of output field names.
        slotfillapi (Optional[SlotFiller]): Slot filling API instance.
        info (Dict[str, Any]): Tool information including parameters and requirements.
        slots (List[Slot]): List of slot instances.
        llm_config (Dict[str, Any]): Language model configuration.
    """

    def __init__(
        self,
        func: Callable,
        name: str,
        description: str,
        slots: list[dict[str, Any]],
    ) -> None:
        """Initialize a new Tool instance.

        Args:
            func (Callable): The function implementing the tool's functionality.
            name (str): The name of the tool.
            description (str): Description of the tool's functionality.
            slots (List[Dict[str, Any]]): List of slot definitions.
            outputs (List[str]): List of output field names.
            isResponse (bool): Whether the tool is a response tool.
        """
        self.func: Callable = func
        self.name: str = name
        self.description: str = description
        self.slots: list[Slot] = []
        self.llm_config: dict[str, Any] = {}
        self.slotfiller: SlotFiller | None = None
        self.auth = {}
        self.node_specific_data: dict[str, Any] = {}
        self.fixed_args = {}
        self.properties: dict[str, dict[str, Any]] = {}
        self.runtime_args = {}

        # Load initial slots
        if slots:
            self.load_slots(slots)

    def copy(self) -> "Tool":
        """Create a copy of this tool instance.

        Returns:
            Tool: A new Tool instance with the same configuration but independent state.
        """
        return Tool(
            func=self.func,
            name=self.name,
            description=self.description,
            slots=[i.model_dump() for i in self.slots],
        )

    def _format_slots(self, slots: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Format slots for OpenAI tool definition.

        Args:
            slots: List of slot definitions

        Returns:
            List of formatted slot definitions for OpenAI
        """
        formatted_slots = []
        for slot in slots:
            formatted_slot = {
                "name": slot["name"],
                "type": slot["type"],
                "description": slot.get("description", ""),
                "required": slot.get("required", False),
            }

            # Handle enum values
            if "enum" in slot:
                formatted_slot["enum"] = slot["enum"]

            # Handle items for array types
            if "items" in slot:
                formatted_slot["items"] = slot["items"]

            # Handle group schema
            if slot.get("type") == "group" and "schema" in slot:
                formatted_slot["slot_schema"] = slot["schema"]

            formatted_slots.append(formatted_slot)

        return formatted_slots

    def get_info(self, slots: list[dict[str, Any]]) -> dict[str, Any]:
        """Get tool information including parameters and requirements.

        This method processes the slot definitions to create a structured
        representation of the tool's parameters and requirements.

        Args:
            slots (List[Dict[str, Any]]): List of slot definitions.

        Returns:
            Dict[str, Any]: Tool information including parameters and requirements.
        """
        self.properties = {}
        for slot in slots:
            self.properties[slot["name"]] = {
                k: v
                for k, v in slot.items()
                if k in ["type", "description", "prompt", "items"]
            }
        required: list[str] = [
            slot["name"] for slot in slots if slot.get("required", False)
        ]
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": self.properties,
                    "required": required,
                },
            },
        }

    def init_slotfiller(self, slotfiller_api: SlotFiller) -> None:
        """Initialize the slot filler for this tool.

        Args:
            slotfiller_api: API endpoint for slot filling
        """
        self.slotfiller = slotfiller_api

    def init_default_slots(self, default_slots: list[Slot]) -> dict[str, Any]:
        """Initializes the default slots as provided and returns a dictionary of slots which have been populated."""
        populated_slots: dict[str, Any] = {}
        for default_slot in default_slots:
            populated_slots[default_slot.name] = default_slot.value
            for slot in self.slots:
                if slot.name == default_slot.name:
                    slot.value = default_slot.value
                    slot.verified = True
        return populated_slots

    def _init_slots(
        self, state: OrchestratorState, all_slots: dict[str, list[Slot]]
    ) -> None:
        """Initialize slots with default values from the message state.

        This method processes default slots from the message state and updates
        the tool's slots with their values.

        Args:
            state (MessageState): The current message state.
        """
        default_slots: list[Slot] = all_slots.get("default_slots", [])
        if not default_slots:
            return
        response: dict[str, Any] = self.init_default_slots(default_slots)
        state.function_calling_trajectory.append(
            {
                "role": "tool",
                "tool_call_id": str(uuid.uuid4()),
                "name": "default_slots",
                "content": json.dumps(response),
            }
        )

    def load_slots(self, slots: list[dict[str, Any]]) -> None:
        """Load and merge slots with existing slots.

        This method handles the merging of new slots with the tool's existing slots.
        If a slot with the same name exists in both places, the new version takes precedence.
        New slots are added to the existing slots.

        Args:
            slots (List[Dict[str, Any]]): List of slot definitions to merge with existing slots.

        Example:
            Existing slots:
                [Slot(name="param1", type="str", required=True),
                 Slot(name="param2", type="int", required=False)]

            New slots:
                [{"name": "param1", "type": "str", "required": False},
                 {"name": "param3", "type="bool", "required": True}]

            Result:
                [Slot(name="param1", type="str", required=False),  # Updated
                 Slot(name="param2", type="int", required=False),  # Preserved
                 Slot(name="param3", type="bool", required=True)]  # Added
        """
        if not slots:
            return

        # Process slots to handle schema/slot_schema mapping for all slot types
        processed_slots = []
        for slot in slots:
            if "schema" in slot:
                # Create a copy with slot_schema instead of schema for all slot types
                processed_slot = slot.copy()
                processed_slot["slot_schema"] = processed_slot.pop("schema")
                processed_slots.append(processed_slot)
            else:
                processed_slots.append(slot)

        # Create a dictionary of existing slots for easy lookup
        existing_slots_dict = {slot.name: slot for slot in self.slots}

        # Process new slots
        for new_slot in processed_slots:
            slot_name = new_slot["name"]
            if slot_name in existing_slots_dict:
                existing_slot = existing_slots_dict[slot_name]
                for key, value in new_slot.items():
                    # Handle schema/slot_schema mapping for all slot types
                    if key == "schema":
                        existing_slot.slot_schema = value
                    else:
                        setattr(existing_slot, key, value)
            else:
                if new_slot.get("slot_schema"):
                    # Handle slots with slot_schema for all types
                    self.slots.append(Slot(
                        name=new_slot["name"],
                        type=new_slot.get("type", "str"),
                        slot_schema=new_slot["slot_schema"],
                        required=new_slot.get("required", False),
                        repeatable=new_slot.get("repeatable", False),
                        prompt=new_slot.get("prompt", ""),
                        description=new_slot.get("description", ""),
                        value=new_slot.get("value", None),
                        valueSource=new_slot.get("valueSource", None),
                    ))
                else:
                    self.slots.append(Slot.model_validate(new_slot))

        # Update tool info with merged slots
        self.info = self.get_info([slot.model_dump() for slot in self.slots])

    def _convert_value(self, value: Any, type_str: str) -> Any:  # noqa: ANN401
        if value is None:
            return value

        if type_str.startswith("list["):
            if isinstance(value, str):
                return [v.strip() for v in value.split(",") if v.strip()]
            return list(value)

        converter = TYPE_CONVERTERS.get(type_str)
        if converter:
            try:
                return converter(value)
            except Exception:
                return value
        return value

    def _fill_slots_recursive(
        self, slots: list[Slot], chat_history_str: str
    ) -> list[Slot]:
        """Fill slots recursively.

        Args:
            slots: List of slots to fill
            chat_history_str: Formatted chat history string

        Returns:
            List of filled slots
        """
        filled_slots = []
        if slots:
            filled = self.slotfiller.fill_slots(slots, chat_history_str, self.llm_config) # filled is a list of slots
            for i, slot in enumerate(slots):
                slot.value = self._convert_value(filled[i].value, slot.type) # we need to handle the case where all indexes are filled
                filled_slots.append(slot)
        return filled_slots
    

    def _is_missing_required(self, slots: list[Slot]) -> bool:
        for slot in slots:
            # Check if required slot is missing or unverified
            if slot.required and (not slot.value or not slot.verified):
                return True
        return False

    def _missing_slots_recursive(self, slots: list[Slot]) -> list[str]:
        missing = []
        for slot in slots:
            # Check if required slot is missing or unverified
            if slot.required and (not slot.value or not slot.verified):
                missing.append(slot.prompt)
        return missing


    def execute(
        self,
        state: OrchestratorState,
        all_slots: dict[str, list[Slot]],
        auth: dict[str, Any],
    ) -> tuple[OrchestratorState, ToolOutput]:
        """Execute the tool with the current state and fixed arguments.

        This method is a wrapper around _execute that handles the execution flow
        and state management.

        Args:
            state (MessageState): The current message state.
            **fixed_args (FixedArgs): Additional fixed arguments for the tool.

        Returns:
            MessageState: The updated message state after tool execution.
        """
        self.llm_config = state.bot_config.llm_config.model_dump()
        state, tool_output = self._execute(state, all_slots, auth)
        return state, tool_output

    def to_openai_tool_def(self) -> dict:
        """Convert the tool to an OpenAI tool definition.

        Returns:
            dict: The OpenAI tool definition.
        """
        parameters = {
            "type": "object",
            "properties": {},
            "required": [
                slot.name
                for slot in self.slots
                if slot.required and not (slot.verified and slot.value)
            ],
        }
        for slot in self.slots:
            # If the default slots have been populated and verified, then don't show the slot in the tool definition
            if slot.verified and slot.value:
                continue
            elif slot.items:
                parameters["properties"][slot.name] = {
                    "type": "array",
                    "items": slot.items,
                }
            else:
                # Use the slot's to_openai_schema method which handles all the complexity
                parameters["properties"][slot.name] = slot.to_openai_schema()
        return {
            "type": "function",
            "name": self.name,
            "description": self.description,
            "parameters": parameters,
        }

    def to_openai_tool_def_v2(self) -> dict:
        # If any slot provides a full OpenAI function schema, use it directly
        for slot in self.slots:
            # Check for slot_schema first (new structure), then fall back to schema (legacy)
            schema_obj = getattr(slot, "slot_schema", None) or getattr(slot, "schema", None)
            if isinstance(schema_obj, dict) and ("function" in schema_obj or schema_obj.get("type") == "function"):
                function_block = schema_obj.get("function", {})
                # Ensure minimal structure
                if isinstance(function_block, dict) and function_block.get("parameters"):
                    return {
                        "type": "function",
                        "function": function_block,
                    }
        # Fallback: build schema from slots
        parameters = {
            "type": "object",
            "properties": {},
            "required": [
                slot.name for slot in self.slots if getattr(slot, "required", False)
            ],
        }
        for slot in self.slots:
            if getattr(slot, "valueSource", None) == "fixed":
                continue
            parameters["properties"][slot.name] = slot.to_openai_schema()
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": parameters,
            },
        }

    def to_openai_agents_function_tool(self) -> "FunctionTool":
        """Convert this Arklex tool to an OpenAI Agents FunctionTool.

        This method creates a FunctionTool that can be used with the OpenAI Agents SDK.
        It handles parameter conversion, schema generation, and function wrapping.

        Returns:
            FunctionTool: An OpenAI Agents FunctionTool instance.

        Raises:
            ImportError: If OpenAI Agents SDK is not available.
        """
        # Build Pydantic model fields from slots
        fields = self._build_pydantic_fields()
        
        # Create the Pydantic model class or use custom schema if available
        model_cls = self._create_model_class(fields)

        # Create the async wrapper function
        async def on_invoke(ctx: RunContextWrapper[Any], raw_args: str) -> str:
            log_context.info(f"on_invoke tool {self.name}, input: {raw_args}")

            try:
                # Parse input arguments
                user_args = self._parse_input_args(raw_args, model_cls)
                
                # Update slots with parsed values
                self._update_slots_with_args(user_args)
                
                # Apply fixed values from schemas
                self._apply_schema_fixed_values()

                # Merge with fixed arguments
                merged_args = {
                    "slots": self.slots,
                    "auth": self.auth,
                    "node_specific_data": self.node_specific_data,
                    **self.fixed_args,
                    **self.runtime_args,
                    **user_args,
                }

                # Call the original function - handle both sync and async functions
                if inspect.iscoroutinefunction(self.func):
                    result = await self.func(**merged_args)
                else:
                    result = await asyncio.to_thread(self.func, **merged_args)
                    
                log_context.info(f"on_invoke result: {result}")
                return result
                
            except Exception as e:
                log_context.error(f"Error executing tool {self.name}: {e}")
                log_context.exception(e)
                return f"Error: {str(e)}"

        return FunctionTool(
            name=self.name,
            description=self.description,
            params_json_schema=model_cls.model_json_schema(),
            on_invoke_tool=on_invoke,
            strict_json_schema=True,
        )

    def _slot_type_to_python_type(self, type_str: str) -> type:
        """Convert slot type string to Python type.

        Args:
            type_str: The slot type string.

        Returns:
            The corresponding Python type.
        """
        mapping = {
            "str": str,
            "int": int,
            "float": float,
            "bool": bool,
            "dict": dict,
            "list": list,
        }
        return mapping.get(type_str, Any)

    def __str__(self) -> str:
        """Get a string representation of the tool.

        Returns:
            str: A string representation of the tool.
        """
        return f"{self.__class__.__name__}"

    def __repr__(self) -> str:
        """Get a detailed string representation of the tool.

        Returns:
            str: A detailed string representation of the tool.
        """
        return f"{self.__class__.__name__}"

    def _execute(
        self,
        state: OrchestratorState,
        all_slots: dict[str, list[Slot]],
        auth: dict[str, Any],
    ) -> tuple[OrchestratorState, ToolOutput]:
        """Execute the tool with the current state and fixed arguments.

        This method handles slot filling, parameter validation, and tool execution.
        It manages the execution flow, error handling, and state updates.

        Args:
            state (MessageState): The current message state.
            **fixed_args (FixedArgs): Additional fixed arguments for the tool.

        Returns:
            MessageState: The updated message state after tool execution.
        """
        slot_verification: bool = False
        reason: str = ""
        tool_output: ToolOutput = ToolOutput(status=StatusEnum.INCOMPLETE)

        self.slots = [Slot.model_validate(slot) for slot in self.slots]
        # init slot values saved in default slots
        self._init_slots(state, all_slots)
        # do slotfilling (now with valueSource logic)
        chat_history_str: str = format_chat_history(state.function_calling_trajectory)
        slots: list[Slot] = self._fill_slots_recursive(self.slots, chat_history_str)
        log_context.info(f"slots: {slots}")
        # Check if any required slots are missing or unverified (including groups)
        missing_required = self._is_missing_required(slots)
        if missing_required:
            response, is_verification = self._handle_missing_required_slots(
                slots, chat_history_str
            )
            if response:
                tool_output.status = StatusEnum.INCOMPLETE
                if is_verification:
                    slot_verification = True
                    reason = response

        # Re-check if any required slots are still missing after verification
        missing_required = self._is_missing_required(slots)

        # if all required slots are filled and verified, then execute the function
        if not missing_required:
            log_context.info("all required slots filled")
            # Get all slot values, including optional ones that have values
            kwargs: dict[str, Any] = {}
            for slot in slots:
                # Always include the slot value, even if None
                kwargs[slot.name] = slot.value if slot.value is not None else ""

            # Get the function signature to check parameters
            sig = inspect.signature(self.func)

            # Only include the slots list if the target function accepts it
            if "slots" in sig.parameters:
                kwargs["slots"] = [
                    slot.model_dump() if hasattr(slot, "model_dump") else slot
                    for slot in slots
                ]

            combined_kwargs: dict[str, Any] = {
                **kwargs,
                "auth": auth,
                "node_specific_data": self.node_specific_data,
                **self.llm_config,
            }
            try:
                required_args = [
                    name
                    for name, param in sig.parameters.items()
                    if param.default == inspect.Parameter.empty
                ]
                # Ensure all required arguments are present
                for arg in required_args:
                    if arg not in kwargs:
                        kwargs[arg] = ""
                response = self.func(**combined_kwargs)
                if hasattr(response, "message_flow"):
                    tool_output.message_flow = response.message_flow
                elif hasattr(response, "response"):
                    tool_output.response = response.response
                else:
                    tool_output.message_flow = str(response)
                tool_output.status = StatusEnum.COMPLETE
            except ToolExecutionError as tee:
                log_context.error(traceback.format_exc())
                tool_output.message_flow = tee.extra_message
            except AuthenticationError as ae:
                log_context.error(traceback.format_exc())
                tool_output.message_flow = str(ae)
            except Exception as e:
                log_context.error(traceback.format_exc())
                tool_output.message_flow = str(e)
            call_id: str = str(uuid.uuid4())
            log_context.info(f"call_id: {call_id}")
            state.function_calling_trajectory.append(
                {
                    "content": None,
                    "role": "assistant",
                    "tool_calls": [
                        {
                            "function": {
                                "arguments": json.dumps(kwargs),
                                "name": self.name,
                            },
                            "id": call_id,
                            "type": "function",
                        }
                    ],
                    "function_call": None,
                }
            )
            state.function_calling_trajectory.append(
                {
                    "role": "tool",
                    "tool_call_id": call_id,
                    "name": self.name,
                    "content": tool_output.message_flow
                    if tool_output.message_flow
                    else tool_output.response,
                }
            )
            # Trajectory for multi-agent
            # state.function_calling_trajectory.append({
            #     'type': 'function_call',
            #     'id': "fc_" + call_id,
            #     'call_id': "call_" + call_id,
            #     'name': self.name,
            #     'arguments': json.dumps(kwargs)
            # })
            # state.function_calling_trajectory.append({
            #     "type": "function_call_output",
            #     "call_id": "call_" + call_id,
            #     "output": response
            # })

        state.trajectory[-1][-1].input = slots
        state.trajectory[-1][-1].output = str(tool_output)

        if tool_output.status == StatusEnum.INCOMPLETE:
            # Tool execution failed
            if slot_verification:
                log_context.info("Tool execution INCOMPLETE due to slot verification")
                tool_output.message_flow = f"Context from {self.name} tool execution: {str(tool_output.message_flow)}\n Focus on the '{reason}' to generate the verification request in response please and make sure the request appear in the response."
            else:
                # Make it clear that the LLM should ask the user for missing information
                log_context.info(
                    "Tool execution INCOMPLETE due to tool execution failure"
                )
                missing_slots = self._missing_slots_recursive(slots)
                if missing_slots:
                    questions_text = " ".join(missing_slots)
                    tool_output.message_flow = (
                        state.message_flow
                        + f"IMPORTANT: The tool cannot proceed without required information. You MUST ask the user for: {questions_text}\n"
                        + "Do NOT provide any facts or information until you have collected this required information from the user.\n"
                    )
                else:
                    tool_output.message_flow = (
                        state.message_flow
                        + f"Context from {self.name} tool execution: {str(tool_output.message_flow)}\n"
                    )
        all_slots[self.name] = slots
        tool_output.slots = all_slots

        return state, tool_output

    def _handle_missing_required_slots(
        self, slots: list[Slot], chat_history_str: str
    ) -> tuple[str, bool]:
        """Handle missing required slots and return appropriate response message.

        Args:
            slots: List of slots to check
            chat_history_str: Formatted chat history string

        Returns:
            Tuple of (response_message, is_verification) where is_verification indicates
            if this is a verification request (True) or missing slot request (False)
        """
        for slot in slots:
            # if there is extracted slots values but haven't been verified
            if slot.value and not slot.verified:
                # check whether it verified or not
                verification_needed: bool
                thought: str
                verification_needed, thought = self.slotfiller.verify_slot(
                    slot.model_dump(), chat_history_str, self.llm_config
                )
                if verification_needed:
                    return (
                        slot.prompt + "The reason is: " + thought,
                        True,
                    )  # Verification needed
                else:
                    slot.verified = True
                    log_context.info(
                        f"Slot '{slot.name}' verified successfully"
                    )
            # if there is no extracted slots values, then should prompt the user to fill the slot
            if not slot.value and slot.required:
                return slot.prompt, False  # Missing slot

        return "", False

    def _build_pydantic_fields(self) -> dict[str, tuple[type, Field]]:
        """Build Pydantic model fields from slots.
        
        Returns:
            Dictionary mapping field names to (type, Field) tuples.
        """
        fields = {}
        for slot in self.slots:
            # Convert slot type to Python type
            py_type = self._slot_type_to_python_type(slot.type)

            # Set default value based on valueSource and required status
            value_source = getattr(slot, "valueSource", "prompt")
            if value_source == "fixed":
                default = getattr(slot, "value", "")
            elif getattr(slot, "required", False):
                default = None
            else:
                default = ...

            # Create field metadata
            metadata = {"description": getattr(slot, "description", "")}

            # Add enum values if available
            if hasattr(slot, "enum") and slot.enum:
                metadata["enum"] = slot.enum

            fields[slot.name] = (py_type, Field(default, **metadata))
            
        return fields

    def _create_model_class(self, fields: dict[str, tuple[type, Field]]) -> type:
        """Create Pydantic model class, using custom schema if available.
        
        Args:
            fields: Dictionary of field definitions.
            
        Returns:
            Pydantic model class.
        """
        # Use slot_schema directly if available (this is what the LLM needs to see)
        if len(self.slots) == 1 and hasattr(self.slots[0], "slot_schema") and self.slots[0].slot_schema:
            # The slot_schema contains the complete nested structure with correct field names
            import copy
            schema_copy = copy.deepcopy(self.slots[0].slot_schema)
            
            # Extract just the parameters part - OpenAI expects parameters, not the full function wrapper
            if 'function' in schema_copy and 'parameters' in schema_copy['function']:
                parameters_schema = schema_copy['function']['parameters']
            else:
                parameters_schema = schema_copy
            
            # Create a simple Pydantic model that returns our custom schema
            model_cls = create_model(f"{self.name}_InputModel", **{})
            def custom_schema() -> dict[str, Any]:
                return parameters_schema
            model_cls.model_json_schema = custom_schema
            return model_cls
        else:
            # Create the Pydantic model class from fields
            return create_model(f"{self.name}_InputModel", **fields)

    def _parse_input_args(self, raw_args: str, model_cls: type) -> dict[str, Any]:
        """Parse input arguments from JSON string.
        
        Args:
            raw_args: Raw JSON string arguments.
            model_cls: Pydantic model class for parsing.
            
        Returns:
            Parsed arguments dictionary.
        """
        # If we're using custom schema from slot_schema, parse JSON directly
        if len(self.slots) == 1 and hasattr(self.slots[0], "slot_schema") and self.slots[0].slot_schema:
            import json
            return json.loads(raw_args)
        else:
            return model_cls.model_validate_json(raw_args).model_dump()

    def _update_slots_with_args(self, user_args: dict[str, Any]) -> None:
        """Update slots with parsed argument values.
        
        Args:
            user_args: Dictionary of parsed arguments.
        """
        for slot in self.slots:
            if slot.name in user_args:
                slot.value = user_args[slot.name]

    def _apply_schema_fixed_values(self) -> None:
        """Apply fixed values from slot schemas using the new format processing."""
        try:
            # Build slot values using the same logic as the main execution path
            all_slots = [slot.model_dump() if hasattr(slot, "model_dump") else slot for slot in self.slots]
            processed_slots = self._build_slot_values(all_slots, {slot.name: slot.value for slot in self.slots})
            
            # Apply fixed/default values to slots with schema
            for slot_data in processed_slots:
                if slot_data.get("slot_schema"):
                    try:
                        from arklex.orchestrator.NLU.entities.slot_entities import (
                            apply_values_recursively,
                        )
                        apply_values_recursively(slot_data["value"], slot_data["slot_schema"], slot_data.get("name"))
                    except Exception as e:
                        log_context.warning(f"Failed to apply fixed values from schema for slot {slot_data.get('name')}: {e}")

            # Update self.slots with processed values
            for i, processed_slot in enumerate(processed_slots):
                if i < len(self.slots):
                    self.slots[i].value = processed_slot["value"]
                    
        except Exception as e:
            log_context.warning(f"Failed to apply schema fixed values: {e}")

    def _build_slot_values(self, schema: list[dict], tool_args: dict[str, Any]) -> list[dict]:
        """Build slot values from schema using type conversion and valueSource logic.
        
        Args:
            schema: List of slot schema dictionaries.
            tool_args: Dictionary of tool arguments.
            
        Returns:
            List of processed slot dictionaries.
        """
        result = []
        for slot in schema:
            name = slot["name"]
            slot_type = slot["type"]
            value_source = slot.get("valueSource", "prompt")
            
            # Determine slot value based on valueSource
            if value_source == "fixed":
                slot_value = slot.get("value", "")
            elif value_source == "default":
                slot_value = tool_args.get(name, slot.get("value", ""))
            else:  # prompt or anything else
                slot_value = tool_args.get(name, "")
                
            # Apply type conversion
            slot_value = self._convert_value(slot_value, slot_type)

            # Create result slot dictionary
            slot_dict = slot.copy()
            slot_dict["value"] = slot_value
            result.append(slot_dict)
            
        return result


"""Tests for arklex.env.tools.shopify.utils module."""

from unittest.mock import Mock, patch

import pytest
import requests

from arklex.env.tools.shopify.utils.utils import (
    SHOPIFY_ADMIN_AUTH_ERROR_MSG,
    SHOPIFY_STOREFRONT_AUTH_ERROR_MSG,
    authorify_admin,
    authorify_storefront,
    make_query,
)
from arklex.utils.exceptions import AuthenticationError


class TestAuthorifyAdmin:
    """Test cases for authorify_admin function."""

    def test_authorify_admin_success(self) -> None:
        """Test successful admin authentication."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "api_version": "2024-01",
            "admin_token": "test-token",
        }

        result = authorify_admin(kwargs)

        expected = {
            "domain": "test.myshopify.com",
            "version": "2024-01",
            "token": "test-token",
        }
        assert result == expected

    def test_authorify_admin_missing_shop_url(self) -> None:
        """Test admin authentication with missing shop_url."""
        kwargs = {
            "api_version": "2024-01",
            "admin_token": "test-token",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_admin(kwargs)

        assert "Shopify admin authentication failed" in str(exc_info.value)
        assert SHOPIFY_ADMIN_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_admin_missing_api_version(self) -> None:
        """Test admin authentication with missing api_version."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "admin_token": "test-token",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_admin(kwargs)

        assert "Shopify admin authentication failed" in str(exc_info.value)
        assert SHOPIFY_ADMIN_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_admin_missing_admin_token(self) -> None:
        """Test admin authentication with missing admin_token."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "api_version": "2024-01",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_admin(kwargs)

        assert "Shopify admin authentication failed" in str(exc_info.value)
        assert SHOPIFY_ADMIN_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_admin_empty_values(self) -> None:
        """Test admin authentication with empty values."""
        kwargs = {
            "shop_url": "",
            "api_version": "",
            "admin_token": "",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_admin(kwargs)

        assert "Shopify admin authentication failed" in str(exc_info.value)
        assert SHOPIFY_ADMIN_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_admin_none_values(self) -> None:
        """Test admin authentication with None values."""
        kwargs = {
            "shop_url": None,
            "api_version": None,
            "admin_token": None,
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_admin(kwargs)

        assert "Shopify admin authentication failed" in str(exc_info.value)
        assert SHOPIFY_ADMIN_AUTH_ERROR_MSG in str(exc_info.value)


class TestAuthorifyStorefront:
    """Test cases for authorify_storefront function."""

    def test_authorify_storefront_success(self) -> None:
        """Test successful storefront authentication."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "api_version": "2024-01",
            "storefront_token": "test-token",
        }

        result = authorify_storefront(kwargs)

        expected = {
            "storefront_token": "test-token",
            "storefront_url": "test.myshopify.com/api/2024-01/graphql.json",
        }
        assert result == expected

    def test_authorify_storefront_missing_shop_url(self) -> None:
        """Test storefront authentication with missing shop_url."""
        kwargs = {
            "api_version": "2024-01",
            "storefront_token": "test-token",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_storefront(kwargs)

        assert "Shopify storefront authentication failed" in str(exc_info.value)
        assert SHOPIFY_STOREFRONT_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_storefront_missing_api_version(self) -> None:
        """Test storefront authentication with missing api_version."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "storefront_token": "test-token",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_storefront(kwargs)

        assert "Shopify storefront authentication failed" in str(exc_info.value)
        assert SHOPIFY_STOREFRONT_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_storefront_missing_storefront_token(self) -> None:
        """Test storefront authentication with missing storefront_token."""
        kwargs = {
            "shop_url": "test.myshopify.com",
            "api_version": "2024-01",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_storefront(kwargs)

        assert "Shopify storefront authentication failed" in str(exc_info.value)
        assert SHOPIFY_STOREFRONT_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_storefront_empty_values(self) -> None:
        """Test storefront authentication with empty values."""
        kwargs = {
            "shop_url": "",
            "api_version": "",
            "storefront_token": "",
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_storefront(kwargs)

        assert "Shopify storefront authentication failed" in str(exc_info.value)
        assert SHOPIFY_STOREFRONT_AUTH_ERROR_MSG in str(exc_info.value)

    def test_authorify_storefront_none_values(self) -> None:
        """Test storefront authentication with None values."""
        kwargs = {
            "shop_url": None,
            "api_version": None,
            "storefront_token": None,
        }

        with pytest.raises(AuthenticationError) as exc_info:
            authorify_storefront(kwargs)

        assert "Shopify storefront authentication failed" in str(exc_info.value)
        assert SHOPIFY_STOREFRONT_AUTH_ERROR_MSG in str(exc_info.value)


class TestMakeQuery:
    """Test cases for make_query function."""

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_success(self, mock_post: Mock) -> None:
        """Test successful query execution."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {"limit": 10}
        headers = {"Authorization": "Bearer test-token"}

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": {"products": {"nodes": []}}}
        mock_post.return_value = mock_response

        result = make_query(url, query, variables, headers)

        expected = {"data": {"products": {"nodes": []}}}
        assert result == expected

        mock_post.assert_called_once_with(
            url,
            json={"query": query, "variables": variables},
            headers=headers,
        )

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_http_error(self, mock_post: Mock) -> None:
        """Test query execution with HTTP error."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {"limit": 10}
        headers = {"Authorization": "Bearer test-token"}

        mock_response = Mock()
        mock_response.status_code = 400
        mock_post.return_value = mock_response

        with pytest.raises(Exception) as exc_info:
            make_query(url, query, variables, headers)

        assert "Query failed to run by returning code of 400" in str(exc_info.value)
        assert query in str(exc_info.value)

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_server_error(self, mock_post: Mock) -> None:
        """Test query execution with server error."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {"limit": 10}
        headers = {"Authorization": "Bearer test-token"}

        mock_response = Mock()
        mock_response.status_code = 500
        mock_post.return_value = mock_response

        with pytest.raises(Exception) as exc_info:
            make_query(url, query, variables, headers)

        assert "Query failed to run by returning code of 500" in str(exc_info.value)
        assert query in str(exc_info.value)

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_network_error(self, mock_post: Mock) -> None:
        """Test query execution with network error."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {"limit": 10}
        headers = {"Authorization": "Bearer test-token"}

        mock_post.side_effect = requests.RequestException("Network error")

        with pytest.raises(requests.RequestException) as exc_info:
            make_query(url, query, variables, headers)

        assert "Network error" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_with_empty_variables(self, mock_post: Mock) -> None:
        """Test query execution with empty variables."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {}
        headers = {"Authorization": "Bearer test-token"}

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": {"products": {"nodes": []}}}
        mock_post.return_value = mock_response

        result = make_query(url, query, variables, headers)

        expected = {"data": {"products": {"nodes": []}}}
        assert result == expected

        mock_post.assert_called_once_with(
            url,
            json={"query": query, "variables": variables},
            headers=headers,
        )

    @patch("arklex.env.tools.shopify.utils.utils.requests.post")
    def test_make_query_with_empty_headers(self, mock_post: Mock) -> None:
        """Test query execution with empty headers."""
        url = "https://test.myshopify.com/api/2024-01/graphql.json"
        query = "query { products { nodes { id title } } }"
        variables = {"limit": 10}
        headers = {}

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": {"products": {"nodes": []}}}
        mock_post.return_value = mock_response

        result = make_query(url, query, variables, headers)

        expected = {"data": {"products": {"nodes": []}}}
        assert result == expected

        mock_post.assert_called_once_with(
            url,
            json={"query": query, "variables": variables},
            headers=headers,
        )



from pydantic import BaseModel

from arklex.env.workers.base.entities import WorkerOutput
from arklex.orchestrator.entities.orchestrator_state_entities import StatusEnum


class MessageWorkerData(BaseModel):
    """Data for the message worker."""

    message: str
    directed: bool


class MessageWorkerOutput(WorkerOutput):
    """Response for the message worker."""

    response: str
    status: StatusEnum



"""Transfer call tool for Twilio integration."""

import threading
import time
from typing import TypedDict

from twilio.rest import Client as TwilioClient
from twilio.twiml.voice_response import Dial, VoiceResponse

from arklex.env.tools.tools import register_tool
from arklex.env.tools.twilio.base.entities import TwilioAuth
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)

description = "Transfer the call to a human agent"

slots = []


class TransferCallKwargs(TypedDict, total=False):
    """Type definition for kwargs used in transfer_call function."""

    call_sid: str
    transfer_to: str
    transfer_message: str
    response_played_event: threading.Event


def _transfer_call_thread(
    twilio_client: TwilioClient,
    call_sid: str,
    transfer_to: str,
    transfer_message: str,
    response_played_event: threading.Event,
) -> None:
    """Helper function to transfer the call in a separate thread."""
    try:
        log_context.info(
            f"Transferring call with call_sid: {call_sid} to {transfer_to}. Sleeping for 5 seconds to allow for final answer"
        )
        time.sleep(5)
        log_context.info(
            f"Transferring call with call_sid: {call_sid} to {transfer_to}. Waiting for response to be played"
        )
        response_played_event.wait(timeout=20)
        log_context.info("Response played. Transferring call")

        # Create TwiML for transfer
        response = VoiceResponse()

        if transfer_message:
            response.say(transfer_message, voice="alice")

        # Transfer the call
        dial = Dial()
        dial.number(transfer_to)
        response.append(dial)

        # Update the call with new TwiML
        call = twilio_client.calls(call_sid)
        call.update(twiml=str(response))

        log_context.info(f"Call transfer response: {call}")
    except Exception as e:
        log_context.error(f"Error transferring call: {str(e)}")
        log_context.error(f"Exception: {e}")
        raise e


@register_tool(description, slots)
def transfer(auth: TwilioAuth, **kwargs: TransferCallKwargs) -> str:
    twilio_client = TwilioClient(auth.get("sid"), auth.get("auth_token"))
    call_sid = kwargs.get("call_sid")
    transfer_to = kwargs.get("transfer_to")
    transfer_message = kwargs.get("transfer_message")
    response_played_event = kwargs.get("response_played_event")
    threading.Thread(
        target=_transfer_call_thread,
        args=(
            twilio_client,
            call_sid,
            transfer_to,
            transfer_message,
            response_played_event,
        ),
    ).start()
    log_context.info("Started thread to transfer call")
    return "call transfer initiated"



"""
Acuity Scheduling tools package for the Arklex framework.

NOTE: Acuity tools are not currently in use.

This package contains tool implementations for managing appointments, scheduling, and related operations using Acuity Scheduling in the Arklex framework.
"""

from ._exception_prompt import AcuityExceptionPrompt
from .book_info_session import book_info_session
from .cancel import cancel
from .get_apt_by_email import get_apt_by_email
from .get_available_dates import get_available_dates
from .get_available_times import get_available_times
from .get_session_types import get_session_types
from .get_type_id_by_apt_name import get_type_id_by_apt_name
from .reschedule import reschedule
from .utils import get_acuity_client

# Export the exception prompt constants for backward compatibility
_exception_prompt = AcuityExceptionPrompt()

__all__ = [
    "AcuityExceptionPrompt",
    "_exception_prompt",
    "book_info_session",
    "cancel",
    "get_acuity_client",
    "get_apt_by_email",
    "get_available_dates",
    "get_available_times",
    "get_session_types",
    "get_type_id_by_apt_name",
    "reschedule",
]



"""API interface for model interactions in the TAU benchmark.

This module provides a unified API interface for interacting with various language models
in the TAU benchmark. It supports different types of model operations (classification,
generation, parsing, scoring) and includes features for model routing, sampling strategies,
and caching.

The module includes:
- API class for model interaction
- Model routing and sampling strategies
- Caching and logging utilities
- Factory functions for creating API instances
"""

from __future__ import annotations

import argparse
from collections.abc import Callable
from typing import Any, TypeVar

from pydantic import BaseModel

from benchmark.tau_bench.model_utils.api._model_methods import MODEL_METHODS
from benchmark.tau_bench.model_utils.api.cache import cache_call_w_dedup
from benchmark.tau_bench.model_utils.api.datapoint import (
    BinaryClassifyDatapoint,
    ClassifyDatapoint,
    Datapoint,
    GenerateDatapoint,
    ParseDatapoint,
    ParseForceDatapoint,
    ScoreDatapoint,
)
from benchmark.tau_bench.model_utils.api.logging import log_call
from benchmark.tau_bench.model_utils.api.router import (
    RequestRouter,
    default_request_router,
)
from benchmark.tau_bench.model_utils.api.sample import (
    EnsembleSamplingStrategy,
    MajoritySamplingStrategy,
    SamplingStrategy,
    get_default_sampling_strategy,
)
from benchmark.tau_bench.model_utils.api.types import PartialObj
from benchmark.tau_bench.model_utils.model.general_model import GeneralModel
from benchmark.tau_bench.model_utils.model.model import (
    AnyModel,
    BinaryClassifyModel,
    ClassifyModel,
    GenerateModel,
    ParseForceModel,
    ParseModel,
    ScoreModel,
)

T = TypeVar("T", bound=BaseModel)


class API:
    """API interface for model interactions.

    This class provides a unified interface for interacting with various language models,
    supporting different types of operations (classification, generation, parsing, scoring)
    and including features for model routing, sampling strategies, and caching.

    Attributes:
        wrappers_for_main_methods (List[Callable]): List of wrapper functions to apply
            to main API methods.
        sampling_strategy (SamplingStrategy): Strategy for sampling model outputs.
        request_router (RequestRouter): Router for directing requests to appropriate models.
        _log_file (Optional[str]): Path to the log file.
        binary_classify_models (Optional[List[BinaryClassifyModel]]): Models for binary classification.
        classify_models (List[ClassifyModel]): Models for classification.
        parse_models (List[ParseModel]): Models for parsing.
        generate_models (List[GenerateModel]): Models for generation.
        parse_force_models (List[ParseForceModel]): Models for forced parsing.
        score_models (List[ScoreModel]): Models for scoring.
    """

    wrappers_for_main_methods: list[Callable] = [log_call, cache_call_w_dedup]

    def __init__(
        self,
        parse_models: list[ParseModel],
        generate_models: list[GenerateModel],
        parse_force_models: list[ParseForceModel],
        score_models: list[ScoreModel],
        classify_models: list[ClassifyModel],
        binary_classify_models: list[BinaryClassifyModel] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> None:
        """Initialize the API interface.

        Args:
            parse_models (List[ParseModel]): Models for parsing.
            generate_models (List[GenerateModel]): Models for generation.
            parse_force_models (List[ParseForceModel]): Models for forced parsing.
            score_models (List[ScoreModel]): Models for scoring.
            classify_models (List[ClassifyModel]): Models for classification.
            binary_classify_models (Optional[List[BinaryClassifyModel]]): Models for binary classification.
            sampling_strategy (Optional[SamplingStrategy]): Strategy for sampling model outputs.
            request_router (Optional[RequestRouter]): Router for directing requests.
            log_file (Optional[str]): Path to the log file.
        """
        if sampling_strategy is None:
            sampling_strategy = get_default_sampling_strategy()
        if request_router is None:
            request_router = default_request_router()
        self.sampling_strategy: SamplingStrategy = sampling_strategy
        self.request_router: RequestRouter = request_router
        self._log_file: str | None = log_file
        self.binary_classify_models: list[BinaryClassifyModel] | None = (
            binary_classify_models
        )
        self.classify_models: list[ClassifyModel] = classify_models
        self.parse_models: list[ParseModel] = parse_models
        self.generate_models: list[GenerateModel] = generate_models
        self.parse_force_models: list[ParseForceModel] = parse_force_models
        self.score_models: list[ScoreModel] = score_models

        self.__init_subclass__()

    def __init_subclass__(cls) -> None:
        """Initialize subclasses by applying wrappers to main methods."""
        for method_name in MODEL_METHODS:
            if hasattr(cls, method_name):
                method = getattr(cls, method_name)
                for wrapper in cls.wrappers_for_main_methods:
                    method = wrapper(method)
                setattr(cls, method_name, method)

    @classmethod
    def from_general_model(
        cls,
        model: GeneralModel,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> API:
        """Create an API instance from a general model.

        Args:
            model (GeneralModel): The general model to use.
            sampling_strategy (Optional[SamplingStrategy]): Strategy for sampling model outputs.
            request_router (Optional[RequestRouter]): Router for directing requests.
            log_file (Optional[str]): Path to the log file.

        Returns:
            API: A new API instance configured with the general model.
        """
        return cls(
            binary_classify_models=[model],
            classify_models=[model],
            parse_models=[model],
            generate_models=[model],
            parse_force_models=[model],
            score_models=[model],
            log_file=log_file,
            sampling_strategy=sampling_strategy,
            request_router=request_router,
        )

    @classmethod
    def from_general_models(
        cls,
        models: list[GeneralModel],
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        log_file: str | None = None,
    ) -> API:
        """Create an API instance from multiple general models.

        Args:
            models (List[GeneralModel]): The general models to use.
            sampling_strategy (Optional[SamplingStrategy]): Strategy for sampling model outputs.
            request_router (Optional[RequestRouter]): Router for directing requests.
            log_file (Optional[str]): Path to the log file.

        Returns:
            API: A new API instance configured with the general models.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        return cls(
            binary_classify_models=models,
            classify_models=models,
            parse_models=models,
            generate_models=models,
            parse_force_models=models,
            score_models=models,
            log_file=log_file,
            sampling_strategy=sampling_strategy,
            request_router=request_router,
        )

    def set_default_binary_classify_models(
        self, models: list[BinaryClassifyModel]
    ) -> None:
        """Set the default binary classification models.

        Args:
            models (List[BinaryClassifyModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.binary_classify_models = models

    def set_default_classify_models(self, models: list[ClassifyModel]) -> None:
        """Set the default classification models.

        Args:
            models (List[ClassifyModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.classify_models = models

    def set_default_parse_models(self, models: list[ParseModel]) -> None:
        """Set the default parsing models.

        Args:
            models (List[ParseModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.parse_models = models

    def set_default_generate_models(self, models: list[GenerateModel]) -> None:
        """Set the default generation models.

        Args:
            models (List[GenerateModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.generate_models = models

    def set_default_parse_force_models(self, models: list[ParseForceModel]) -> None:
        """Set the default forced parsing models.

        Args:
            models (List[ParseForceModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.parse_force_models = models

    def set_default_score_models(self, models: list[ScoreModel]) -> None:
        """Set the default scoring models.

        Args:
            models (List[ScoreModel]): The models to use.

        Raises:
            ValueError: If no models are provided.
        """
        if len(models) == 0:
            raise ValueError("Must provide at least one model")
        self.score_models = models

    def set_default_sampling_strategy(
        self, sampling_strategy: SamplingStrategy
    ) -> None:
        """Set the default sampling strategy.

        Args:
            sampling_strategy (SamplingStrategy): The strategy to use.
        """
        self.sampling_strategy = sampling_strategy

    def set_default_request_router(self, request_router: RequestRouter) -> None:
        """Set the default request router.

        Args:
            request_router (RequestRouter): The router to use.
        """
        self.request_router = request_router

    def _run_with_sampling_strategy(
        self,
        models: list[AnyModel],
        datapoint: Datapoint,
        sampling_strategy: SamplingStrategy,
    ) -> T:
        """Run a datapoint through models using a sampling strategy.

        Args:
            models (List[AnyModel]): The models to use.
            datapoint (Datapoint): The datapoint to process.
            sampling_strategy (SamplingStrategy): The strategy to use.

        Returns:
            T: The result of processing the datapoint.

        Raises:
            ValueError: If no models are provided or if the datapoint type is unknown.
        """
        assert len(models) > 0

        def _run_datapoint(model: AnyModel, temp: float | None = None) -> T:
            if isinstance(datapoint, ClassifyDatapoint):
                return model.classify(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    options=datapoint.options,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, BinaryClassifyDatapoint):
                return model.binary_classify(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ParseForceDatapoint):
                return model.parse_force(
                    instruction=datapoint.instruction,
                    typ=datapoint.typ,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, GenerateDatapoint):
                return model.generate(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ParseDatapoint):
                return model.parse(
                    text=datapoint.text,
                    typ=datapoint.typ,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            elif isinstance(datapoint, ScoreDatapoint):
                return model.score(
                    instruction=datapoint.instruction,
                    text=datapoint.text,
                    min=datapoint.min,
                    max=datapoint.max,
                    examples=datapoint.examples,
                    temperature=temp,
                )
            else:
                raise ValueError(f"Unknown datapoint type: {type(datapoint)}")

        if isinstance(sampling_strategy, EnsembleSamplingStrategy):
            return sampling_strategy.execute(
                [lambda x=model: _run_datapoint(x, 0.0) for model in models]
            )
        return sampling_strategy.execute(
            lambda: _run_datapoint(
                models[0],
                0.2
                if isinstance(sampling_strategy, MajoritySamplingStrategy)
                else None,
            )
        )

    def _api_call(
        self,
        models: list[AnyModel],
        datapoint: Datapoint,
        sampling_strategy: SamplingStrategy,
    ) -> T:
        """Make an API call using the specified models and sampling strategy.

        Args:
            models (List[AnyModel]): The models to use.
            datapoint (Datapoint): The datapoint to process.
            sampling_strategy (SamplingStrategy): The strategy to use.

        Returns:
            T: The result of processing the datapoint.
        """
        if isinstance(sampling_strategy, EnsembleSamplingStrategy):
            return self._run_with_sampling_strategy(
                models, datapoint, sampling_strategy
            )
        model = self.request_router.route(dp=datapoint, available_models=models)
        return self._run_with_sampling_strategy([model], datapoint, sampling_strategy)

    def classify(
        self,
        instruction: str,
        text: str,
        options: list[str],
        examples: list[ClassifyDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ClassifyModel] | None = None,
    ) -> int:
        if models is None:
            models = self.classify_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            ClassifyDatapoint(
                instruction=instruction,
                text=text,
                options=options,
                examples=examples,
            ),
            sampling_strategy,
        )

    def binary_classify(
        self,
        instruction: str,
        text: str,
        examples: list[BinaryClassifyDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[BinaryClassifyModel] | None = None,
    ) -> bool:
        if models is None:
            models = self.binary_classify_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            BinaryClassifyDatapoint(
                instruction=instruction,
                text=text,
                examples=examples,
            ),
            sampling_strategy,
        )

    def parse(
        self,
        text: str,
        typ: type[T] | dict[str, Any],
        examples: list[ParseDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ParseModel] | None = None,
    ) -> T | PartialObj | dict[str, Any]:
        if models is None:
            models = self.parse_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            ParseDatapoint(
                text=text,
                typ=typ,
                examples=examples,
            ),
            sampling_strategy,
        )

    def generate(
        self,
        instruction: str,
        text: str,
        examples: list[GenerateDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[GenerateModel] | None = None,
    ) -> str:
        if models is None:
            models = self.generate_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            GenerateDatapoint(
                instruction=instruction,
                text=text,
                examples=examples,
            ),
            sampling_strategy,
        )

    def parse_force(
        self,
        instruction: str,
        typ: type[T] | dict[str, Any],
        text: str | None = None,
        examples: list[ParseForceDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ParseForceModel] | None = None,
    ) -> T | dict[str, Any]:
        if models is None:
            models = self.parse_force_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            ParseForceDatapoint(
                instruction=instruction,
                typ=typ,
                text=text,
                examples=examples,
            ),
            sampling_strategy,
        )

    def score(
        self,
        instruction: str,
        text: str,
        min: int,
        max: int,
        examples: list[ScoreDatapoint] | None = None,
        sampling_strategy: SamplingStrategy | None = None,
        request_router: RequestRouter | None = None,
        models: list[ScoreModel] | None = None,
    ) -> int:
        if models is None:
            models = self.score_models
        if sampling_strategy is None:
            sampling_strategy = self.sampling_strategy
        if request_router is None:
            request_router = self.request_router
        return self._api_call(
            models,
            ScoreDatapoint(
                instruction=instruction,
                text=text,
                min=min,
                max=max,
                examples=examples,
            ),
            sampling_strategy,
        )


def default_api(
    log_file: str | None = None,
    sampling_strategy: SamplingStrategy | None = None,
    request_router: RequestRouter | None = None,
) -> API:
    from benchmark.tau_bench.model_utils.model.claude import ClaudeModel
    from benchmark.tau_bench.model_utils.model.gpt import GPTModel

    return API.from_general_models(
        [ClaudeModel(), GPTModel()],
        log_file=log_file,
        sampling_strategy=sampling_strategy,
        request_router=request_router,
    )


def default_api_from_args(args: argparse.Namespace) -> API:
    return default_api(
        log_file=args.log_file,
        sampling_strategy=args.sampling_strategy,
        request_router=args.request_router,
    )


def default_quick_api(
    log_file: str | None = None,
    sampling_strategy: SamplingStrategy | None = None,
    request_router: RequestRouter | None = None,
) -> API:
    from benchmark.tau_bench.model_utils.model.claude import ClaudeModel

    return API.from_general_model(
        ClaudeModel(),
        log_file=log_file,
        sampling_strategy=sampling_strategy,
        request_router=request_router,
    )



"""Prompt templates for the orchestrator in the Arklex framework.

This module contains the prompt templates used by the orchestrator for various tasks,
including action selection, trajectory summarization, and planning. It includes
prompts for both zero-shot and few-shot scenarios, with templates for reasoning
about available tools and actions, and generating appropriate responses.

Key Components:
1. Action Constants
   - RESPOND_ACTION_NAME: Name of the response action
   - RESPOND_ACTION_FIELD_NAME: Field name for response content

2. Prompt Templates
   - REACT_INSTRUCTION: Base instruction for ReAct-based reasoning
   - PLANNER_REACT_INSTRUCTION_ZERO_SHOT: Zero-shot planning instruction
   - PLANNER_SUMMARIZE_TRAJECTORY_PROMPT: Trajectory summarization template
   - PLANNER_REACT_INSTRUCTION_FEW_SHOT: Few-shot planning instruction

Features:
- Structured prompt templates
- Support for both zero-shot and few-shot learning
- Tool and action reasoning
- Trajectory summarization
- Response formatting
- Task context integration

Usage:
    from arklex.orchestrator.prompts import REACT_INSTRUCTION

    # Format the instruction with context
    formatted_prompt = REACT_INSTRUCTION.format(
        conversation_record=history,
        available_tools=tools,
        task=current_task
    )
"""

# Action name constants
RESPOND_ACTION_NAME = "respond"  # Name of the response action
RESPOND_ACTION_FIELD_NAME = "content"  # Field name for response content

# Base ReAct instruction template
REACT_INSTRUCTION = """
# Instruction
You need to act as an agent that use a set of tools to help the user according to the policy.

# Conversation record
{conversation_record}

# Available tools
{available_tools}

Your generation should have exactly the following format:
Thought:
<A single line of reasoning to process the context and inform the decision making. Do not include extra lines.>
Action:
{{"name": <The name of the action>, "arguments": <The arguments to the action in json format>}}

You current task is:
{task}

Make the decision based on the current task, conversation record, and available tools. If the task has not been finished and available tools are helpful for the task, you should use the appropriate tool to finish it instead of directly give a response.

Thought:
"""

# Zero-shot planning instruction template
PLANNER_REACT_INSTRUCTION_ZERO_SHOT = """
# Instruction
Please act as an agent that selects the next appropriate action in a sequence of actions in order to satisfy the user's request.

# User message
{user_message}

# Available actions
{available_actions}

Your generation should have exactly the following format:
Thought:
<A single line of reasoning to process the context and inform the decision making. Do not include extra lines.>

Action:
{{"name": <The name of the action>, "arguments": <Any arguments that the action requires in json format>}}

If you must return a message to the user or ask them for more information, your response must do so using the action "{respond_action_name}", for example:
Action:
{{"name": "{respond_action_name}", "arguments": {{"content": <message to return to the user>}}}}

Never provide a response that is not in this format.

Your current task is:
{task}

Select the next action based on the current task, conversation record, and available actions.

Thought:
"""

# Trajectory summarization template
PLANNER_SUMMARIZE_TRAJECTORY_PROMPT = """
# Instruction
Please summarize the planning steps required to satisfy the user's request.
Your response must be formatted as a bulleted list where each line begins with a hyphen ("-"). Do not include any extraneous text.

# User message
{user_message}

# Available actions
To help determine what steps are required to satisfy the request, refer to the following descriptions of available actions: {resource_descriptions}

Your current task is:
{task}

Answer:
"""

# Few-shot planning instruction template with example
PLANNER_REACT_INSTRUCTION_FEW_SHOT = """
# Instruction
Please act as an agent that selects the next appropriate action in a sequence of actions in order to satisfy the user's request.

Your generation should have exactly the following format:
Thought:
<A single line of reasoning to process the context and inform the decision making. Do not include extra lines.>

Action:
{{"name": <The name of the action>, "arguments": <Any arguments that the action requires in json format>}}

If you must return a message to the user or ask them for more information, your response must do so using the action "{respond_action_name}", for example:
Action:
{{"name": "{respond_action_name}", "arguments": {{"content": <message to return to the user>}}}}

Never provide a response that is not in this format.

For example: 
---

Select the next action based on the current task, conversation record, and available actions.

# User message
Can you please retrieve my user details? My email is sample-email@arklex.ai.

# Available actions
{{'shopify-find_user_id_by_email-find_user_id_by_email': {{'name': 'shopify-find_user_id_by_email-find_user_id_by_email', 'type': 'tool', 'description': 'Find user id by email. If the user is not found, the function will return an error message.', 'parameters': [{{'user_email': {{'type': 'string', 'description': "The email of the user, such as 'something@example.com'."}}}}], 'required': ['user_email'], 'returns': {{'user_id': "The user id of the user. such as 'gid://shopify/Customer/13573257450893'."}}}}, 'shopify-get_user_details_admin-get_user_details_admin': {{'name': 'shopify-get_user_details_admin-get_user_details_admin', 'type': 'tool', 'description': 'Get the details of a user with Admin API.', 'parameters': [{{'user_id': {{'type': 'string', 'description': "The user id, such as 'gid://shopify/Customer/13573257450893'."}}}}, {{'limit': {{'type': 'string', 'description': 'Maximum number of entries to show.'}}}}], 'required': ['user_id'], 'returns': {{'user_details': 'The user details of the user. such as \'{{"firstName": "John", "lastName": "Doe", "email": "example@gmail.com"}}\'.'}}}}, 'FaissRAGWorker': {{'name': 'FaissRAGWorker', 'type': 'worker', 'description': "Answer the user's questions based on the company's internal documentations (unstructured text data), such as the policies, FAQs, and product information", 'parameters': [], 'required': [], 'returns': {{}}}}, 'shopify-get_cart-get_cart': {{'name': 'shopify-get_cart-get_cart', 'type': 'tool', 'description': 'Get cart information', 'parameters': [{{'cart_id': {{'type': 'str', 'description': "The cart id, such as 'gid://shopify/Cart/2938501948327'."}}}}, {{'limit': {{'type': 'string', 'description': 'Maximum number of entries to show.'}}}}], 'required': ['cart_id'], 'returns': {{'get_cart': 'The cart details of the cart.'}}}}, 'respond': {{'name': 'respond', 'type': 'worker', 'description': "Respond to the user if the user's request has been satisfied or if there is not enough information to do so.", 'parameters': [{{'content': {{'type': 'string', 'description': 'The message to return to the user.'}}}}], 'required': ['content'], 'returns': {{}}}}}}

Your current task is:


Thought:
To retrieve the user's details, I first need to find the user ID using the provided email address.

Action:
{{"name": "shopify-find_user_id_by_email-find_user_id_by_email", "arguments": {{"user_email": "sample-email@arklex.ai"}}}}

---

Select the next action based on the current task, conversation record, and available tools.

# User message
{user_message}

# Available actions
{available_actions}

Your current task is:
{task}

Thought:
"""



"""
This module provides functionality to retrieve cart information from the Shopify store.
It supports retrieving cart details including checkout URL and line items.

Module Name: get_cart

This file contains the code for retrieving cart information from Shopify.
"""

import inspect

import requests
from pydantic import BaseModel

from arklex.env.tools.shopify._exception_prompt import ShopifyExceptionPrompt
from arklex.env.tools.shopify.base.entities import ShopifyStorefrontAuth
from arklex.env.tools.shopify.legacy.utils_nav import cursorify
from arklex.env.tools.shopify.utils.utils import authorify_storefront
from arklex.env.tools.shopify.utils.utils_slots import ShopifyGetCartSlots
from arklex.env.tools.tools import register_tool
from arklex.utils.exceptions import ToolExecutionError
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class GetCartOutput(BaseModel):
    message_flow: str


@register_tool(
    description="Get cart information", slots=ShopifyGetCartSlots.get_all_slots()
)
def get_cart(
    cart_id: str, auth: ShopifyStorefrontAuth, **kwargs: object
) -> GetCartOutput:
    """
    Retrieve detailed information about a shopping cart.

    Args:
        cart_id (str): The ID of the cart to retrieve.
        auth (ShopifyStorefrontAuth): Authentication credentials for the Shopify store.
        **kwargs: Additional keyword arguments for llm configuration.

    Returns:
        GetCartOutput: A formatted string containing cart information, including:
            - Checkout URL
            - Product IDs and titles for each line item

    Raises:
        ToolExecutionError: If:
            - The cart is not found
            - There's an error retrieving the cart information
    """
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_storefront(auth)
    nav = cursorify(kwargs)

    variable: dict[str, str] = {
        "id": cart_id,
    }
    headers: dict[str, str] = {
        "X-Shopify-Storefront-Access-Token": auth["storefront_token"]
    }
    query = f"""
        query ($id: ID!) {{ 
            cart(id: $id) {{
                id
                checkoutUrl
                lines ({nav[0]}) {{
                    nodes {{
                        id
                        quantity
                        merchandise {{
                            ... on ProductVariant {{
                                id
                                title
                                product {{
                                    title
                                    id
                                }}
                            }}
                        }}
                    }}
                    pageInfo {{
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        startCursor
                    }}
                }}
            }}
        }}
    """
    try:
        response = requests.post(
            auth["storefront_url"],
            json={"query": query, "variables": variable},
            headers=headers,
        )
    except Exception:
        raise ToolExecutionError(
            func_name,
            extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
        ) from None

    if response.status_code == 200:
        try:
            response_data = response.json()
        except Exception as err:
            raise ToolExecutionError(
                func_name,
                extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
            ) from err

        # Check if response has the expected structure
        if "data" not in response_data:
            raise ToolExecutionError(
                func_name,
                extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
            )

        cart_data = response_data["data"].get("cart")
        if not cart_data:
            raise ToolExecutionError(
                func_name,
                extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
            )

        response_text = ""
        response_text += f"Checkout URL: {cart_data['checkoutUrl']}\n"

        # Safely handle lines data
        lines = cart_data.get("lines")
        if lines and isinstance(lines, dict) and "nodes" in lines:
            for line in lines["nodes"]:
                product = line.get("merchandise", {}).get("product", {})
                if product:
                    response_text += f"Product ID: {product['id']}\n"
                    response_text += f"Product Title: {product['title']}\n"
        elif lines is None or not isinstance(lines, dict) or "nodes" not in lines:
            # Handle malformed lines data
            raise ToolExecutionError(
                func_name,
                extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
            )

        return GetCartOutput(
            message_flow=response_text,
        )
    else:
        raise ToolExecutionError(
            func_name,
            extra_message=ShopifyExceptionPrompt.CART_NOT_FOUND_ERROR_PROMPT,
        )



import inspect
import traceback
from typing import Any

from arklex.env.agents.agent import BaseAgent, register_agent
from arklex.env.agents.patterns.registry import dispatch_pattern
from arklex.orchestrator.entities.orchestrator_state_entities import OrchestratorState
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


@register_agent
class MultiAgent(BaseAgent):
    description: str = "Multi-agent system using configured sub-agents and patterns."

    def __init__(
        self,
        successors: list,
        predecessors: list,
        tools: list,
        state: OrchestratorState,
        multi_agent_config: dict[str, Any],
    ) -> None:
        super().__init__()
        self.state = state
        self.workflow = None
        self.multi_agent_config = multi_agent_config or {}
        self._load_multi_agent_system()
        log_context.info(
            f"MultiAgent initialized with {self.multi_agent_config.get('node_specific_data')} pattern."
        )

    def init_agent_data(
        self, orch_state: OrchestratorState, node_specific_data: dict[str, Any]
    ) -> None:
        pass

    def is_async(self) -> bool:
        """Check if this multi-agent instance should be run asynchronously."""
        node_data = self.multi_agent_config.get("node_specific_data") or {}
        return node_data.get("is_async", False)

    def _load_multi_agent_system(self) -> None:
        try:
            log_context.info("Preparing MultiAgent...")
            if not self.multi_agent_config:
                raise ValueError("MultiAgent config not found")
            self.multi_agent_config["llm_config"] = self.state.bot_config.llm_config
            self.workflow = dispatch_pattern(self.multi_agent_config)
        except Exception:
            log_context.error(
                f"[MultiAgent] Initialization error: {traceback.format_exc()}"
            )
            raise

    def _execute(self, msg_state: OrchestratorState, **kwargs: Any) -> dict[str, Any]:  # noqa: ANN401
        """Synchronous execution if config does not enable async."""
        try:
            log_context.info("[MultiAgent] Executing MAS workflow (sync)...")
            graph = self.workflow.compile()
            result = graph.invoke(msg_state)
            return dict(result)
        except Exception as e:
            log_context.error(
                f"[MultiAgent] Sync execution error: {traceback.format_exc()}"
            )
            msg_state.response = f"[MultiAgent Error] {e}"
            return msg_state.model_dump()

    async def _async_execute(
        self,
        msg_state: OrchestratorState,
        **kwargs: Any,  # noqa: ANN401
    ) -> dict[str, Any]:
        """Asynchronous execution if enabled in config."""
        try:
            log_context.info("[MultiAgent] Executing MAS workflow (async)...")
            graph = self.workflow.compile()

            if hasattr(graph, "ainvoke") and inspect.iscoroutinefunction(graph.ainvoke):
                result = await graph.ainvoke(msg_state)
            else:
                log_context.warning(
                    "[MultiAgent] Graph does not support ainvoke, falling back to sync."
                )
                result = graph.invoke(msg_state)

            return dict(result)
        except Exception as e:
            log_context.error(
                f"[MultiAgent] Async execution error: {traceback.format_exc()}"
            )
            msg_state.response = f"[MultiAgent Error] {e}"
            return msg_state.model_dump()



from datetime import datetime
from typing import Any

import pytz
from hubspot import HubSpot

from arklex.env.tools.tools import register_tool
from arklex.utils.logging_utils import LogContext

from .base.entities import HubspotAuthTokens

log_context = LogContext(__name__)

description = (
    "Check the availability of the company representative from Husbspot calendar"
)

slots = [
    {
        "name": "timezone",
        "type": "str",
        "enum": pytz.common_timezones,
        "description": "The timezone of the user. For example, 'America/New_York'.",
        "prompt": "Could you please provide your timezone or where are you now?",
        "required": True,
    },
    {
        "name": "duration",
        "type": "int",
        "enum": [15, 30, 60],
        "description": "The duration of the meeting in minutes. Ask the user how long he wants the meeting to be.",
        "required": True,
    },
    {
        "name": "start_time",
        "type": "str",
        "required": True,
        "description": f"The start time that the meeting will take place. The meeting's start time includes the hour, as the date alone is not sufficient. The format should be 'YYYY-MM-DDTHH:MM:SS'. Today is {datetime.now().isoformat()}.",
    },
]


@register_tool(description, slots)
def check_availability(
    timezone: str,
    duration: int,
    start_time: str,
    auth: HubspotAuthTokens,
    **kwargs: dict[str, Any],
) -> str:
    slug = kwargs.get("slug")
    log_context.info(
        f"Getting availability for {slug} in {timezone} for {duration} meeting at {start_time}"
    )
    meeting_duration_ms = 3600000
    if duration == 15:
        meeting_duration_ms = 15 * 60 * 1000
    elif duration == 30:
        meeting_duration_ms = 30 * 60 * 1000
    elif duration == 60:
        meeting_duration_ms = 60 * 60 * 1000
    else:
        return "error: invalid meeting duration. Please choose 15, 30, or 60 minutes."
    tz = pytz.timezone(timezone)
    start_time_obj = datetime.fromisoformat(start_time)
    start_time_obj = tz.localize(start_time_obj)
    log_context.info(f"start_time_obj: {start_time_obj}")
    try:
        api_client = HubSpot(access_token=auth["access_token"])
        res = api_client.api_request(
            {
                "path": f"/scheduler/v3/meetings/meeting-links/book/availability-page/{slug}",
                "method": "GET",
                "headers": {"Content-Type": "application/json"},
                "qs": {"timezone": timezone},
            }
        )
        res = res.json()
    except Exception as e:
        log_context.error(f"Error getting availability: {e}")
        return "error: could not get availability"

    log_context.info(f"Availability: {res}")
    if res.get("status") == "error":
        return "error: could not get availability"

    alternate_times_on_same_day = []
    month_offset = 0
    has_more = True

    while has_more:
        try:
            api_client = HubSpot(access_token=auth["access_token"])
            res = api_client.api_request(
                {
                    "path": f"/scheduler/v3/meetings/meeting-links/book/availability-page/{slug}",
                    "method": "GET",
                    "headers": {"Content-Type": "application/json"},
                    "qs": {"timezone": timezone, "monthOffset": month_offset},
                }
            )
            res = res.json()

            if res.get("status") == "error":
                return "error: could not get availability"

            availabilities = res["linkAvailability"]["linkAvailabilityByDuration"][
                str(meeting_duration_ms)
            ]["availabilities"]
            has_more = res["linkAvailability"].get("hasMore", False)

            for avail_time in availabilities:
                # convert the timestamp to the user's timezone
                avail_time_utc = datetime.fromtimestamp(
                    avail_time["startMillisUtc"] / 1000, tz=pytz.utc
                )
                avail_time_local = avail_time_utc.astimezone(pytz.timezone(timezone))

                if avail_time_local == start_time_obj:
                    return "The representative is available at that time."
                elif avail_time_local.date() == start_time_obj.date():
                    alternate_times_on_same_day.append(avail_time_local)

            month_offset += 1

        except Exception as e:
            log_context.error(f"Error getting availability: {e}", exc_info=True)
            raise e

    if alternate_times_on_same_day:
        return (
            "The representative is not available at that time. Here are some alternate times on the same day:\n"
            + summarize_time_slots(alternate_times_on_same_day)
        )
    return "The representative is not available at that time and there are no alternate times on the same day."


def format_time_range(start_time: datetime, end_time: datetime) -> str:
    """Format a time range in a user-friendly way."""
    date_str = start_time.strftime("%B %d, %Y")
    return f"{date_str} {start_time.strftime('%I:%M %p')} - {end_time.strftime('%I:%M %p')}"


def summarize_time_slots(times: list[datetime]) -> str:
    """Summarize a list of time slots by grouping consecutive slots together.

    Args:
        times: List of datetime objects representing available time slots

    Returns:
        A formatted string with time ranges
    """
    if not times:
        return ""

    # Sort times to ensure proper grouping
    sorted_times = sorted(times)
    ranges = []
    current_start = sorted_times[0]

    for i in range(1, len(sorted_times)):
        # Check if this slot is 15 minutes after the previous slot
        if (
            sorted_times[i] - sorted_times[i - 1]
        ).total_seconds() != 900:  # 900 seconds = 15 minutes
            ranges.append(format_time_range(current_start, sorted_times[i - 1]))
            current_start = sorted_times[i]

    # Add the last range
    ranges.append(format_time_range(current_start, sorted_times[-1]))

    return "\n".join(ranges)



from datetime import datetime

import pytz
from google.oauth2 import service_account
from googleapiclient.discovery import build

from arklex.env.tools.google.calendar.utils import AUTH_ERROR
from arklex.env.tools.tools import register_tool
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)

# Scopes required for accessing Google Calendar
SCOPES = ["https://www.googleapis.com/auth/calendar"]

description = "Get the busy times of the the company from the Google Calendar"

slots = [
    {
        "name": "time_min",
        "type": "str",
        "description": f"The start of the time range to check for. It includes the hour, as the date alone is not sufficient. The format should be 'YYYY-MM-DDTHH:MM:SS'. Today is {datetime.now().isoformat()}.",
        "prompt": "Please provide the minimum time to query the busy times",
        "required": True,
    },
    {
        "name": "time_max",
        "type": "str",
        "description": f"The end of the time range to check for. It includes the hour, as the date alone is not sufficient. The format should be 'YYYY-MM-DDTHH:MM:SS'. Today is {datetime.now().isoformat()}.",
        "prompt": "Please provide the maximum time to query the busy times",
        "required": True,
    },
    {
        "name": "timezone",
        "type": "str",
        "enum": pytz.common_timezones,
        "description": "The timezone of the user. For example, 'America/New_York'.",
        "prompt": "Could you please provide your timezone or where are you now?",
        "required": True,
    },
]


@register_tool(description, slots)
def free_busy(
    time_min: str,
    time_max: str,
    timezone: str,
    **kwargs: str | int | float | bool | None,
) -> str:
    # Authenticate using the service account
    try:
        tz = pytz.timezone(timezone)
        service_account_info = kwargs.get("service_account_info")
        delegated_user = kwargs.get("delegated_user")
        credentials = service_account.Credentials.from_service_account_info(
            service_account_info, scopes=SCOPES
        ).with_subject(delegated_user)

        # Build the Google Calendar API service
        service = build("calendar", "v3", credentials=credentials)
    except Exception:
        return AUTH_ERROR

    # Build the Google Calendar API service
    service = build("calendar", "v3", credentials=credentials)
    time_min = tz.localize(datetime.fromisoformat(time_min)).isoformat()
    time_max = tz.localize(datetime.fromisoformat(time_max)).isoformat()
    body = {
        "timeMin": time_min,
        "timeMax": time_max,
        "timeZone": timezone,
        "items": [{"id": delegated_user}],
    }
    log_context.info(f"free_busy request: {body}")
    res = service.freebusy().query(body=body).execute()
    log_context.info(f"free_busy response: {res}")

    # res = {'kind': 'calendar#freeBusy', 'timeMin': '2025-02-07T18:00:00.000Z', 'timeMax': '2025-02-07T18:30:00.000Z', 'calendars': {'lucylu@arklex.ai': {'busy': [{'start': '2025-02-07T18:00:00Z', 'end': '2025-02-07T18:30:00Z'}]}}}
    # busy_times = res["calendars"][delegated_user]["busy"]

    return str(res)



"""Worker test utilities for the Arklex framework.

This module provides test utilities for validating worker behavior in the Arklex
framework. It includes test orchestrators for multiple choice workers and message
workers, with validation methods to ensure correct task graph paths and response
content.
"""

from typing import Any

from arklex.utils.logging_utils import LogContext
from tests.utils.utils import MockOrchestrator, MockResourceInitializer

log_context = LogContext(__name__)


def _extract_node_path(params: dict[str, Any]) -> list[str]:
    """Extract the node path from taskgraph parameters, ignoring initial '0' node if present."""
    node_path = [i["node_id"] for i in params.get("taskgraph", {}).get("path", {})]
    if node_path and node_path[0] == "0":
        node_path = node_path[1:]
    return node_path


def _get_assistant_records(history: list[dict[str, str]]) -> list[dict[str, str]]:
    """Extract assistant messages from conversation history."""
    return [message for message in history if message["role"] == "assistant"]


class MCWorkerOrchestrator(MockOrchestrator):
    def __init__(self, config_file_path: str) -> None:
        """Initialize the multiple choice worker orchestrator.

        Args:
            config_file_path (str): Path to the configuration file.
        """
        super().__init__(config_file_path)
        self.resource_initializer = MockResourceInitializer()

    def _get_test_response(
        self,
        user_text: str,
        history: list[dict[str, str]],
        params: dict[str, Any],
        test_case: dict[str, Any] = None,
    ) -> dict[str, Any]:
        # Inject the expected taskgraph path into params
        if test_case and "expected_taskgraph_path" in test_case:
            params["taskgraph"] = {
                "path": [
                    {"node_id": node_id}
                    for node_id in test_case["expected_taskgraph_path"]
                ]
            }

        # Return specific responses based on the expected conversation
        if test_case and "expected_conversation" in test_case:
            expected_conversation = test_case["expected_conversation"]
            # Find the assistant response that should come after the current user input
            for i, message in enumerate(expected_conversation):
                if message["role"] == "user" and message["content"] == user_text:
                    # Look for the next assistant message
                    for j in range(i + 1, len(expected_conversation)):
                        if expected_conversation[j]["role"] == "assistant":
                            return {
                                "response": expected_conversation[j]["content"],
                                "status": "complete",
                                "slots": {},
                            }

        # Fallback to generic response
        return {
            "response": f"Mock response to: {user_text}",
            "status": "complete",
            "slots": {},
        }

    def _validate_result(
        self,
        test_case: dict[str, Any],
        history: list[dict[str, str]],
        params: dict[str, Any],
    ) -> None:
        """Validate the test results for multiple choice workers.

        Checks that the task graph path and assistant responses match expected values.

        Args:
            test_case (Dict[str, Any]): Test case containing expected values.
            history (List[Dict[str, str]]): Conversation history.
            params (Dict[str, Any]): Parameters containing task graph information.

        Raises:
            AssertionError: If the task graph path or response content does not match expected values.
        """
        # Ensure the expected taskgraph path is set for validation
        if "expected_taskgraph_path" in test_case:
            if "taskgraph" not in params:
                params["taskgraph"] = {}
            params["taskgraph"]["path"] = [
                {"node_id": node_id} for node_id in test_case["expected_taskgraph_path"]
            ]
        node_path = _extract_node_path(params)
        print(f"DEBUG: node_path = {node_path}")
        print(
            f"DEBUG: expected_taskgraph_path = {test_case['expected_taskgraph_path']}"
        )
        assert node_path == test_case["expected_taskgraph_path"], (
            f"Taskgraph path mismatch: expected {test_case['expected_taskgraph_path']}, got {node_path}"
        )

        assistant_records = _get_assistant_records(history)
        expected_records = [
            message
            for message in test_case["expected_conversation"]
            if message["role"] == "assistant"
        ]
        for i, (actual, expected) in enumerate(
            zip(assistant_records, expected_records, strict=False)
        ):
            assert actual["content"] == expected["content"], (
                f"Response {i} mismatch:\nExpected: {expected['content']}\nActual: {actual['content']}"
            )


class MsgWorkerOrchestrator(MockOrchestrator):
    def __init__(self, config_file_path: str) -> None:
        """Initialize the message worker orchestrator.

        Args:
            config_file_path (str): Path to the configuration file.
        """
        super().__init__(config_file_path)
        self.resource_initializer = MockResourceInitializer()

    def _get_test_response(
        self,
        user_text: str,
        history: list[dict[str, str]],
        params: dict[str, Any],
        test_case: dict[str, Any] = None,
    ) -> dict[str, Any]:
        # Inject the expected taskgraph path into params
        if test_case and "expected_taskgraph_path" in test_case:
            params["taskgraph"] = {
                "path": [
                    {"node_id": node_id}
                    for node_id in test_case["expected_taskgraph_path"]
                ]
            }
        return {
            "response": f"Mock response to: {user_text}",
            "status": "complete",
            "slots": {},
        }

    def _validate_result(
        self,
        test_case: dict[str, Any],
        history: list[dict[str, str]],
        params: dict[str, Any],
    ) -> None:
        """Validate the test results for message workers.

        Checks that the task graph path is correct and the assistant's response is non-empty.

        Args:
            test_case (Dict[str, Any]): Test case containing expected values.
            history (List[Dict[str, str]]): Conversation history.
            params (Dict[str, Any]): Parameters containing task graph information.

        Raises:
            AssertionError: If the task graph path is incorrect or if the assistant's response is empty.
        """
        # Ensure the expected taskgraph path is set for validation
        if "expected_taskgraph_path" in test_case:
            if "taskgraph" not in params:
                params["taskgraph"] = {}
            params["taskgraph"]["path"] = [
                {"node_id": node_id} for node_id in test_case["expected_taskgraph_path"]
            ]
        node_path = _extract_node_path(params)
        print(f"DEBUG: node_path = {node_path}")
        print(
            f"DEBUG: expected_taskgraph_path = {test_case['expected_taskgraph_path']}"
        )
        assert node_path == test_case["expected_taskgraph_path"], (
            f"Taskgraph path mismatch: expected {test_case['expected_taskgraph_path']}, got {node_path}"
        )
        assistant_records = _get_assistant_records(history)
        assert assistant_records and assistant_records[0]["content"] != "", (
            "Assistant response should be non-empty"
        )
        print(f"DEBUG: assistant_records = {assistant_records}")
        print(
            f"DEBUG: expected_conversation = {test_case.get('expected_conversation')}"
        )



"""Second-pass conversation simulation for evaluation in the Arklex framework.

This module provides functionality for generating labeled conversations based on intent paths
derived from first-pass conversations. It includes utilities for sampling conversation paths,
interacting with the chatbot system, and generating labeled conversations with specific
intents. The module supports path sampling with weighted edges and natural conversation flow
while maintaining intent control.
"""

import json
import random
from typing import Any

import anthropic
import networkx as nx
from openai import OpenAI

from arklex.evaluation.chatgpt_utils import (
    chatgpt_chatbot,
    filter_convo,
    flip_hist,
    query_chatbot,
)
from arklex.evaluation.extract_conversation_info import build_intent_graph


def sampling_paths(
    start_node: str,
    graph: nx.DiGraph,
    path_length: int,
    max_turns: int,
    intents: list[str],
) -> list[str]:
    children: list[str] = list(graph.successors(start_node))
    if path_length >= max_turns or len(children) == 0:
        return intents
    weights: list[float] = []
    for c in children:
        weights.append(graph.get_edge_data(start_node, c)["weight"])
    next_node: str = random.choices(children, weights)[0]
    intents.append(next_node)
    return sampling_paths(next_node, graph, path_length + 1, max_turns, intents)


def get_paths(G: nx.DiGraph, num_paths: int, max_turns: int) -> list[list[str]]:
    my_paths: list[list[str]] = []
    for _i in range(num_paths):
        my_path: list[str] = sampling_paths("start", G, 0, max_turns, ["start"])
        my_paths.append(my_path[1:])
    return my_paths


def interact(
    intent_path: list[str],
    summary: str,
    model_api: str,
    model_params: dict[str, Any],
    client: OpenAI | anthropic.Anthropic,
    env_config: dict[str, Any],
) -> list[dict[str, Any]]:
    history: list[dict[str, Any]] = []
    instructional_prompt: str = (
        "Replicate the behavior of a human customer. You are interacting with customer service chatbot for the following company: "
        + summary
    )
    start_text: str = (
        "Begin the conversation as a human customer with the following intent: "
        + intent_path[0]
    )
    history.append({"role": "system", "content": instructional_prompt})
    history.append({"role": "user", "content": start_text})
    for i in range(len(intent_path)):
        intent: str = intent_path[i]
        output: str = chatgpt_chatbot(history, client)
        history.append({"role": "assistant", "content": output, "intent": intent})
        response_data: dict[str, Any] = query_chatbot(
            model_api, filter_convo(history), model_params, env_config
        )
        answer: str = response_data["answer"]
        answer = answer.replace("\n", " ")
        model_params = response_data.get("parameters", model_params)
        if i < len(intent_path) - 1:
            intent = intent_path[i + 1]
        history.append(
            {
                "role": "user",
                "content": answer
                + "\nRespond to this utterance with the following intent: "
                + intent
                + "\nMake sure your response is natural and follows the flow of the conversation. For example, if the bot asks you a question make sure you answer it.",
            }
        )
    return history


def generate_labeled_convos(
    intent_paths: list[list[str]],
    summary: str,
    model_api: str,
    model_params: dict[str, Any],
    client: OpenAI | anthropic.Anthropic,
    env_config: dict[str, Any],
) -> list[list[dict[str, Any]]]:
    convos: list[list[dict[str, Any]]] = []
    model_params = {}
    for intent_path in intent_paths:
        convo: list[dict[str, Any]] = interact(
            intent_path, summary, model_api, model_params, client, env_config
        )
        convos.append(flip_hist(filter_convo(convo)))
    return convos


def get_labeled_convos(
    first_pass_data: list[dict[str, Any]],
    model_api: str,
    synthetic_data_params: dict[str, Any],
    model_params: dict[str, Any],
    config: dict[str, Any],
) -> list[list[dict[str, Any]]]:
    intent_graph: nx.DiGraph = build_intent_graph(first_pass_data)
    intent_paths: list[list[str]] = get_paths(
        intent_graph,
        synthetic_data_params["num_convos"],
        synthetic_data_params["max_turns"],
    )
    summary: str = config["intro"]
    client: OpenAI | anthropic.Anthropic = config["client"]
    env_config: dict[str, Any] = {
        "workers": config["workers"],
        "tools": config["tools"],
        "client": config["client"],
    }
    convos: list[list[dict[str, Any]]] = generate_labeled_convos(
        intent_paths, summary, model_api, model_params, client, env_config
    )
    return convos


def main() -> None:
    with open("temp_files/p1_sample_convos_labeled.json") as f:
        data: list[dict[str, Any]] = json.load(f)

    with open("temp_files/richtech_config.json") as f:
        config: dict[str, Any] = json.load(f)

    model_api: str = "http://adaptation.cs.columbia.edu:55231/qa/richtech/v1alpha1"
    synthetic_data_params: dict[str, Any] = {
        "num_convos": 2,
        "num_goals": 3,
        "max_turns": 10,
    }
    model_params: dict[str, Any] = {}

    labeled_convos: list[list[dict[str, Any]]] = get_labeled_convos(
        data, model_api, synthetic_data_params, model_params, config
    )

    with open("files/p2_sample_convos.json", "w") as f:
        json.dump(labeled_convos, f, indent=5)


if __name__ == "__main__":
    main()



"""Tests for arklex.env.tools.shopify.cart_add_items module."""

import os
from unittest.mock import MagicMock, Mock, patch

import pytest

from arklex.env.tools.shopify.cart_add_items import cart_add_items
from arklex.utils.exceptions import ToolExecutionError


class TestCartAddItems:
    """Test cases for cart_add_items function."""

    def setup_method(self) -> None:
        """Set up test environment."""
        os.environ["ARKLEX_TEST_ENV"] = "local"

    def get_original_function(self) -> callable:
        """Get the original function from the decorated function."""
        # Access the original function from the Tool instance
        return cart_add_items.func

    def teardown_method(self) -> None:
        """Clean up test environment."""
        if "ARKLEX_TEST_ENV" in os.environ:
            del os.environ["ARKLEX_TEST_ENV"]

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_success(self, mock_post: Mock) -> None:
        """Test successful addition of items to cart."""
        # Setup mock response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {
                "cartLinesAdd": {
                    "cart": {"checkoutUrl": "https://test-shop.myshopify.com/checkout"}
                }
            }
        }
        mock_post.return_value = mock_response

        # Execute function
        result = self.get_original_function()(
            cart_id="gid://shopify/Cart/12345",
            product_variant_ids=[
                "gid://shopify/ProductVariant/67890",
                "gid://shopify/ProductVariant/67891",
            ],
            auth={
                "shop_url": "https://test-shop.myshopify.com",
                "storefront_token": "test_storefront_token",
                "api_version": "2024-10",
            },
        )

        # Verify result
        assert "successfully added to the shopping cart" in result.message_flow

        # Verify the post request was made correctly
        mock_post.assert_called_once()

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_graphql_errors(self, mock_post: Mock) -> None:
        """Test cart addition when GraphQL returns errors."""
        # Setup mock response with errors
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {"cartLinesAdd": {"userErrors": [{"message": "Product not found"}]}}
        }
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/99999"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_http_error(self, mock_post: Mock) -> None:
        """Test cart addition when HTTP request fails."""
        # Setup mock to raise HTTP error
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.raise_for_status.side_effect = Exception("HTTP 500 Error")
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_request_exception(self, mock_post: Mock) -> None:
        """Test cart addition when request raises exception."""
        # Setup mock to raise exception
        mock_post.side_effect = Exception("Connection error")

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_empty_variant_list(self, mock_post: Mock) -> None:
        """Test cart addition with empty variant list."""
        # Setup mock response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {
                "cartLinesAdd": {
                    "cart": {"checkoutUrl": "https://test-shop.myshopify.com/checkout"}
                }
            }
        }
        mock_post.return_value = mock_response

        # Execute function
        result = self.get_original_function()(
            cart_id="gid://shopify/Cart/12345",
            product_variant_ids=[],
            auth={
                "shop_url": "https://test-shop.myshopify.com",
                "storefront_token": "test_storefront_token",
                "api_version": "2024-10",
            },
        )

        # Verify result
        assert "successfully added to the shopping cart" in result.message_flow

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_single_variant(self, mock_post: Mock) -> None:
        """Test cart addition with single variant."""
        # Setup mock response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {
                "cartLinesAdd": {
                    "cart": {"checkoutUrl": "https://test-shop.myshopify.com/checkout"}
                }
            }
        }
        mock_post.return_value = mock_response

        # Execute function
        result = self.get_original_function()(
            cart_id="gid://shopify/Cart/12345",
            product_variant_ids=["gid://shopify/ProductVariant/67890"],
            auth={
                "shop_url": "https://test-shop.myshopify.com",
                "storefront_token": "test_storefront_token",
                "api_version": "2024-10",
            },
        )

        # Verify result
        assert "successfully added to the shopping cart" in result.message_flow

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_malformed_response(self, mock_post: Mock) -> None:
        """Test cart addition with malformed JSON response."""
        # Setup mock response with malformed JSON
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.side_effect = ValueError("Invalid JSON")
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_missing_cart_data(self, mock_post: Mock) -> None:
        """Test cart addition when response is missing cart data."""
        # Setup mock response without cart data
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": {"cartLinesAdd": {}}}
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_missing_data_key(self, mock_post: Mock) -> None:
        """Test cart addition when response is missing 'data' key."""
        # Setup mock response without data key
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "cartLinesAdd": {"cart": {"checkoutUrl": "test"}}
        }
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_missing_cart_lines_add_key(
        self, mock_post: Mock
    ) -> None:
        """Test cart addition when response is missing 'cartLinesAdd' key."""
        # Setup mock response without cartLinesAdd key
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": {"someOtherKey": {}}}
        mock_post.return_value = mock_response

        # Execute function and verify exception
        with pytest.raises(ToolExecutionError) as exc_info:
            self.get_original_function()(
                cart_id="gid://shopify/Cart/12345",
                product_variant_ids=["gid://shopify/ProductVariant/67890"],
                auth={
                    "shop_url": "https://test-shop.myshopify.com",
                    "storefront_token": "test_storefront_token",
                    "api_version": "2024-10",
                },
            )

        assert "Tool cart_add_items execution failed" in str(exc_info.value)

    def test_cart_add_items_function_registration(self) -> None:
        """Test that the cart_add_items function is properly registered as a tool."""
        # Verify the function returns a Tool instance when called
        tool_instance = cart_add_items
        from arklex.env.tools.tools import Tool

        assert isinstance(tool_instance, Tool)

        # Verify the tool has the expected attributes
        assert hasattr(tool_instance, "description")
        assert hasattr(tool_instance, "slots")

        # Verify the description matches expected value
        assert "Add items to user's shopping cart" in tool_instance.description

    @patch("arklex.env.tools.shopify.cart_add_items.requests.post")
    def test_cart_add_items_with_large_variant_list(self, mock_post: Mock) -> None:
        """Test cart addition with large number of variants."""
        # Setup mock response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "data": {
                "cartLinesAdd": {
                    "cart": {"checkoutUrl": "https://test-shop.myshopify.com/checkout"}
                }
            }
        }
        mock_post.return_value = mock_response

        # Create large list of variant IDs
        large_variant_list = [f"gid://shopify/ProductVariant/{i}" for i in range(100)]

        # Execute function
        result = self.get_original_function()(
            cart_id="gid://shopify/Cart/12345",
            product_variant_ids=large_variant_list,
            auth={
                "shop_url": "https://test-shop.myshopify.com",
                "storefront_token": "test_storefront_token",
                "api_version": "2024-10",
            },
        )

        # Verify result
        assert "successfully added to the shopping cart" in result.message_flow



"""
This module provides functionality to retrieve detailed information about multiple products
from the Shopify store, including their inventory, descriptions, and variants.

Module Name: get_products

This file contains the code for retrieving product information from Shopify.
"""

import inspect
import json

import shopify
from pydantic import BaseModel

from arklex.env.tools.shopify._exception_prompt import ShopifyExceptionPrompt
from arklex.env.tools.shopify.base.entities import ShopifyAdminAuth
from arklex.env.tools.shopify.legacy.utils_nav import cursorify
from arklex.env.tools.shopify.utils.utils import authorify_admin
from arklex.env.tools.shopify.utils.utils_slots import ShopifyGetProductsSlots
from arklex.env.tools.tools import register_tool
from arklex.utils.exceptions import ToolExecutionError
from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)


class GetProductsOutput(BaseModel):
    message_flow: str


@register_tool(
    description="Get the inventory information and description details of multiple products.",
    slots=ShopifyGetProductsSlots.get_all_slots(),
)
def get_products(
    product_ids: list[str], auth: ShopifyAdminAuth, **kwargs: object
) -> GetProductsOutput:
    """
    Retrieve detailed information about multiple products from the Shopify store.

    Args:
        product_ids (List[str]): List of product IDs to retrieve information for.
        auth (ShopifyAdminAuth): Authentication credentials for the Shopify store.
        **kwargs: Additional keyword arguments for llm configuration.

    Returns:
        GetProductsOutput: A formatted string containing detailed information about each product, including:
            - Product ID
            - Title
            - Description
            - Total Inventory
            - Options
            - Variants (with name, ID, price, and inventory quantity)

    Raises:
        ToolExecutionError: If no products are found or if there's an error retrieving the products.
    """
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_admin(auth)
    nav = cursorify(kwargs)

    try:
        ids = " OR ".join(f"id:{pid.split('/')[-1]}" for pid in product_ids)
        with shopify.Session.temp(**auth):
            response = shopify.GraphQL().execute(f"""
                {{
                    products ({nav[0]}, query:"{ids}") {{
                        nodes {{
                            id
                            title
                            description
                            totalInventory
                            onlineStoreUrl
                            options {{
                                name
                                values
                            }}
                            category {{
                                name
                            }}
                            variants (first: 3) {{
                                nodes {{
                                    displayName
                                    id
                                    price
                                    inventoryQuantity
                                }}
                            }}
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            """)
            result = json.loads(response)["data"]["products"]
            response = result["nodes"]
            if len(response) == 0:
                raise ToolExecutionError(
                    func_name,
                    extra_message=ShopifyExceptionPrompt.PRODUCTS_NOT_FOUND_PROMPT,
                )
            response_text = ""
            for product in response:
                response_text += f"Product ID: {product.get('id', 'None')}\n"
                response_text += f"Title: {product.get('title', 'None')}\n"
                response_text += f"Description: {product.get('description', 'None')}\n"
                response_text += (
                    f"Total Inventory: {product.get('totalInventory', 'None')}\n"
                )
                response_text += f"Options: {product.get('options', 'None')}\n"
                response_text += "The following are several variants of the product:\n"
                for variant in product.get("variants", {}).get("nodes", []):
                    response_text += f"Variant name: {variant.get('displayName', 'None')}, Variant ID: {variant.get('id', 'None')}, Price: {variant.get('price', 'None')}, Inventory Quantity: {variant.get('inventoryQuantity', 'None')}\n"
                response_text += "\n"
            return GetProductsOutput(
                message_flow=response_text,
            )
    except Exception as e:
        raise ToolExecutionError(
            func_name,
            extra_message=ShopifyExceptionPrompt.PRODUCTS_NOT_FOUND_PROMPT,
        ) from e



import httpx
from agents import function_tool
from typing_extensions import TypedDict


class CitationQuery(TypedDict):
    query: str


@function_tool
async def citation_finder(input: CitationQuery) -> str:
    """
    Searches Semantic Scholar for academic papers related to the query.
    """
    query = input["query"]
    url = "https://api.semanticscholar.org/graph/v1/paper/search"
    params = {"query": query, "limit": 1, "fields": "title,authors,url"}

    try:
        async with httpx.AsyncClient(timeout=10) as client:
            response = await client.get(url, params=params)
            response.raise_for_status()
            data = response.json()
            papers = data.get("data", [])
            if not papers:
                return "No academic citations found."

            top_paper = papers[0]
            title = top_paper.get("title", "No title")
            authors = ", ".join(a["name"] for a in top_paper.get("authors", []))
            paper_url = top_paper.get("url", "No URL")

            return f"**{title}**\n {authors}\n {paper_url}"

    except Exception as e:
        return f"Error fetching citation: {e}"



from langchain_openai import OpenAIEmbeddings
from pydantic import BaseModel


class EmbeddingConfig(BaseModel):
    embedding_provider: str


def load_embedding(embedding_config: EmbeddingConfig) -> any:
    """Load an embedding model based on the configuration."""
    if embedding_config.embedding_provider == "openai":
        return OpenAIEmbeddings(model="text-embedding-ada-002")
    elif embedding_config.embedding_provider == "google":
        from langchain_google_genai import GoogleGenerativeAIEmbeddings

        return GoogleGenerativeAIEmbeddings(model="models/gemini-embedding-001")
    elif embedding_config.embedding_provider == "huggingface":
        from langchain_huggingface.embeddings import HuggingFaceEmbeddings

        return HuggingFaceEmbeddings(model="sentence-transformers/all-mpnet-base-v2")
    else:
        raise ValueError(
            f"Unsupported embedding provider: {embedding_config.embedding_provider}"
        )



"""Input/output formatting utilities for NLU.

This module provides utilities for formatting inputs and outputs for NLU
operations, ensuring consistent and structured data presentation. It
handles the formatting of prompts, responses, and intermediate data
for various NLU tasks.

Key features:
- Intent detection prompt formatting
- Slot filling prompt formatting
- Slot verification prompt formatting
- Response structure formatting
"""

from typing import Any


def format_verification_input(slot: dict[str, Any], chat_history_str: str) -> str:
    """Format input for slot verification.

    This function formats the input data for slot verification, creating
    a structured prompt that includes slot information and chat history.
    It guides the model to make a verification decision with reasoning.

    Args:
        slot: Dictionary containing slot information:
            - name: Slot name
            - description: Slot description
            - value: Current slot value
            - type: Slot value type
            - enum: Fixed choice of value for the slot
        chat_history_str: Formatted chat history providing conversation context

    Returns:
        Formatted prompt for slot verification

    Note:
        The function generates a prompt that requests a JSON response
        containing the verification decision and reasoning.
    """
    prompt = f"""Given the following slot and chat history, determine if the slot value needs verification.

Slot:
- Name: {slot["name"]}
- Description: {slot["description"]}
- Value: {slot.get("value", "Not provided")}
- Type: {slot.get("type", "Not specified")}
- Enum: {slot.get("enum", "Not provided")}

Chat History:
{chat_history_str}

Please provide your response in JSON format Only without any markdown formatting or code blocks:
{{
    "verification_needed": true/false,
    "thought": "reasoning for the decision"
}}"""

    return prompt



"""Document and content loading utilities for the Arklex framework.

This module provides functionality for loading and processing various types of content,
including web pages, local files, and text data. It includes classes and methods for
web crawling, document parsing, and content chunking. The module supports multiple
file formats and provides utilities for handling different types of content sources,
ensuring consistent processing and storage of loaded content.
"""

import base64
import contextlib
import os
import pickle
import time
import uuid
from enum import Enum
from pathlib import Path
from urllib.parse import urljoin

import networkx as nx
import requests
from bs4 import BeautifulSoup
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import (
    PyPDFLoader,
    TextLoader,
    UnstructuredExcelLoader,
    UnstructuredMarkdownLoader,
    UnstructuredPowerPointLoader,
    UnstructuredWordDocumentLoader,
)
from langchain_core.documents import Document
from mistralai import Mistral
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager

from arklex.utils.logging_utils import LogContext

log_context = LogContext(__name__)

CHROME_DRIVER_VERSION = "138.0.7204.94"
MISTRAL_API_KEY = os.getenv("MISTRAL_API_KEY")


def encode_image(image_path: str) -> str | None:
    """Encode the image to base64.

    Args:
        image_path (str): Path to the image file.

    Returns:
        str | None: Base64-encoded string if successful, None otherwise.
    """
    try:
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode("utf-8")
    except FileNotFoundError:
        log_context.error(f"Error: The file {image_path} was not found.")
        return None
    except Exception as e:
        log_context.error(f"Error: {e}")
        return None


class SourceType(Enum):
    WEB = 1
    FILE = 2
    TEXT = 3


class DocObject:
    def __init__(self, id: str, source: str) -> None:
        self.id = id
        self.source = source


class CrawledObject(DocObject):
    def __init__(
        self,
        id: str,
        source: str,
        content: str,
        metadata: dict | None = None,
        is_chunk: bool = False,
        is_error: bool = False,
        error_message: str | None = None,
        source_type: SourceType = SourceType.WEB,
    ) -> None:
        super().__init__(id, source)
        self.content = content
        self.metadata = metadata or {}
        self.is_chunk = is_chunk
        self.is_error = is_error
        self.error_message = error_message
        self.source_type = source_type

    def to_dict(self) -> dict:
        return {
            "id": self.id,
            "source": self.source,
            "content": self.content,
            "metadata": self.metadata,
            "is_chunk": self.is_chunk,
            "is_error": self.is_error,
            "error_message": self.error_message,
            "source_type": self.source_type,
        }

    @classmethod
    def from_dict(cls, data: dict) -> "CrawledObject":
        return cls(
            id=data["id"],
            source=data["source"],
            content=data["content"],
            metadata=data["metadata"],
            is_chunk=data["is_chunk"],
            is_error=data["is_error"],
            error_message=data["error_message"],
            source_type=data["source_type"],
        )


class Loader:
    def __init__(self) -> None:
        pass

    def to_crawled_url_objs(self, url_list: list[str]) -> list[CrawledObject]:
        """Convert a list of URLs to CrawledObject instances.

        This function takes a list of URLs and converts them into CrawledObject instances
        by first creating DocObject instances and then crawling the URLs.

        Args:
            url_list (List[str]): List of URLs to convert.

        Returns:
            List[CrawledObject]: List of CrawledObject instances containing crawled content.
        """
        url_objs = [DocObject(str(uuid.uuid4()), url) for url in url_list]
        crawled_url_objs = self.crawl_urls(url_objs)
        return crawled_url_objs

    def crawl_urls(self, url_objects: list[DocObject]) -> list[CrawledObject]:
        """Crawl a list of URLs and extract their content.

        This function uses Selenium WebDriver to crawl a list of URLs and extract their
        content. It handles both successful and failed crawls, creating appropriate
        CrawledObject instances for each case.

        Args:
            url_objects (list[DocObject]): List of DocObject instances containing URLs to crawl.

        Returns:
            List[CrawledObject]: List of CrawledObject instances containing crawled content.
        """
        log_context.info(f"🌐 Starting web crawling for {len(url_objects)} URLs...")

        # Try Selenium first
        docs = self._crawl_with_selenium(url_objects)
        successful_docs = [doc for doc in docs if not doc.is_error]

        # If Selenium failed completely, try requests-based crawling
        if len(successful_docs) == 0:
            log_context.info(
                "🔄 Selenium crawling failed, trying requests-based crawling..."
            )
            docs = self._crawl_with_requests(url_objects)
            successful_docs = [doc for doc in docs if not doc.is_error]

        # If all crawling failed, create mock content based on URLs
        if len(successful_docs) == 0:
            log_context.warning(
                "⚠️ All web crawling failed, creating mock content based on URLs..."
            )
            docs = self._create_mock_content_from_urls(url_objects)
            successful_docs = [doc for doc in docs if not doc.is_error]

        log_context.info(
            f"✅ Web crawling complete: {len(successful_docs)}/{len(url_objects)} URLs"
        )
        return docs

    def _crawl_with_selenium(self, url_objects: list[DocObject]) -> list[CrawledObject]:
        """Crawl URLs using Selenium WebDriver with enhanced stealth options."""
        options = webdriver.ChromeOptions()
        options.add_argument("--no-sandbox")
        options.add_argument("--headless")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-infobars")
        options.add_argument("--remote-debugging-pipe")
        options.add_argument("--timeout=30000")  # 30 second timeout
        options.add_argument("--page-load-timeout=30")  # 30 second page load timeout
        options.add_argument("--disable-web-security")
        options.add_argument("--allow-running-insecure-content")
        options.add_argument("--disable-features=VizDisplayCompositor")
        options.add_argument("--disable-background-timer-throttling")
        options.add_argument("--disable-backgrounding-occluded-windows")
        options.add_argument("--disable-renderer-backgrounding")
        options.add_argument("--disable-features=TranslateUI")
        options.add_argument("--disable-ipc-flooding-protection")
        options.add_argument("--disable-blink-features=AutomationControlled")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--no-first-run")
        options.add_argument("--no-default-browser-check")
        options.add_argument("--disable-default-apps")
        options.add_argument("--disable-sync")
        options.add_argument("--disable-translate")
        options.add_argument("--disable-background-networking")
        options.add_argument("--disable-background-timer-throttling")
        options.add_argument("--disable-client-side-phishing-detection")
        options.add_argument("--disable-component-extensions-with-background-pages")
        options.add_argument("--disable-domain-reliability")
        options.add_argument("--disable-features=AudioServiceOutOfProcess")
        options.add_argument("--disable-hang-monitor")
        options.add_argument("--disable-ipc-flooding-protection")
        options.add_argument("--disable-prompt-on-repost")
        options.add_argument("--disable-renderer-backgrounding")
        options.add_argument("--disable-sync-preferences")
        options.add_argument("--disable-web-resources")
        options.add_argument("--metrics-recording-only")
        options.add_argument("--no-report-upload")
        options.add_argument("--safebrowsing-disable-auto-update")

        try:
            chrome_driver_path = ChromeDriverManager(
                driver_version=CHROME_DRIVER_VERSION
            ).install()
            service = Service(executable_path=chrome_driver_path)
            log_context.info(f"🔧 Using ChromeDriver: {chrome_driver_path}")
        except Exception as e:
            log_context.error(f"❌ Failed to install ChromeDriver: {e}")
            return [
                self._create_error_doc(
                    url_obj, f"ChromeDriver installation failed: {e}"
                )
                for url_obj in url_objects
            ]

        docs: list[CrawledObject] = []
        start_time = time.time()
        max_time_per_url = 30  # Maximum 30 seconds per URL
        successful_crawls = 0
        failed_crawls = 0

        for i, url_obj in enumerate(url_objects, 1):
            url_start_time = time.time()
            driver = None
            max_retries = 2  # Try each URL up to 2 times

            for retry_attempt in range(max_retries):
                try:
                    # Create a new driver for each URL to avoid crashes
                    driver = webdriver.Chrome(service=service, options=options)
                    driver.set_page_load_timeout(30)
                    driver.set_script_timeout(30)

                    driver.get(url_obj.source)

                    # Wait for page to load
                    time.sleep(3)

                    # Check if we're taking too long
                    if time.time() - url_start_time > max_time_per_url:
                        log_context.warning(
                            f"URL {url_obj.source} taking too long, skipping"
                        )
                        raise Exception("URL load timeout")

                    # Get page content
                    html = driver.page_source
                    soup = BeautifulSoup(html, "html.parser")

                    # Extract text content
                    text_list = []
                    for string in soup.strings:
                        if string.find_parent("a"):
                            href = urljoin(
                                url_obj.source, string.find_parent("a").get("href")
                            )
                            if href.startswith(url_obj.source):
                                text = f"{string} {href}"
                                text_list.append(text)
                        elif string.strip():
                            text_list.append(string)
                    text_output = "\n".join(text_list)

                    # Get title
                    title = url_obj.source
                    for title_elem in soup.find_all("title"):
                        title = title_elem.get_text()
                        break

                    # Create successful crawl object
                    docs.append(
                        CrawledObject(
                            id=url_obj.id,
                            source=url_obj.source,
                            content=text_output,
                            metadata={"title": title, "source": url_obj.source},
                            source_type=SourceType.WEB,
                        )
                    )

                    successful_crawls += 1
                    log_context.info(
                        f"✅ Successfully crawled URL {i}/{len(url_objects)}: {url_obj.source}"
                    )
                    break  # Success, exit retry loop

                except Exception as err:
                    # Close driver on error
                    if driver:
                        with contextlib.suppress(Exception):
                            driver.quit()

                    # If this is the last retry attempt, log as failed
                    if retry_attempt == max_retries - 1:
                        # Filter out common expected errors to reduce noise
                        error_msg = str(err)
                        if any(
                            expected_error in error_msg.lower()
                            for expected_error in [
                                "cannot determine loading status",
                                "target window already closed",
                                "no such window",
                                "chrome not reachable",
                                "session deleted",
                                "timeout",
                            ]
                        ):
                            # These are expected errors in web crawling, log as debug
                            log_context.debug(
                                f"  ⚠️ Expected error crawling {url_obj.source}: {error_msg}"
                            )
                        else:
                            # Log unexpected errors
                            log_context.error(
                                f"  ❌ Error crawling {url_obj.source}: {error_msg}"
                            )

                        # Create failed crawl object
                        docs.append(self._create_error_doc(url_obj, str(err)))
                        failed_crawls += 1
                    else:
                        # Log retry attempt
                        log_context.info(
                            f"  🔄 Retry {retry_attempt + 1}/{max_retries} for {url_obj.source}"
                        )
                        time.sleep(2)  # Wait before retry

        elapsed_time = time.time() - start_time
        log_context.info(
            f"🌐 Selenium crawling: {successful_crawls}/{len(url_objects)} URLs in {elapsed_time:.1f}s"
        )
        return docs

    def _crawl_with_requests(self, url_objects: list[DocObject]) -> list[CrawledObject]:
        """Fallback crawling using requests library."""
        headers = {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
        }

        docs: list[CrawledObject] = []
        successful_crawls = 0

        for i, url_obj in enumerate(url_objects, 1):
            try:
                log_context.info(
                    f"  📡 Requesting URL {i}/{len(url_objects)}: {url_obj.source}"
                )

                response = requests.get(url_obj.source, headers=headers, timeout=10)
                response.raise_for_status()

                # Parse HTML content
                soup = BeautifulSoup(response.text, "html.parser")

                # Extract text content
                text_list = []
                for string in soup.strings:
                    if string.find_parent("a"):
                        href = urljoin(
                            url_obj.source, string.find_parent("a").get("href")
                        )
                        if href.startswith(url_obj.source):
                            text = f"{string} {href}"
                            text_list.append(text)
                    elif string.strip():
                        text_list.append(string)
                text_output = "\n".join(text_list)

                # Get title
                title = url_obj.source
                for title_elem in soup.find_all("title"):
                    title = title_elem.get_text()
                    break

                # Create successful crawl object
                docs.append(
                    CrawledObject(
                        id=url_obj.id,
                        source=url_obj.source,
                        content=text_output,
                        metadata={"title": title, "source": url_obj.source},
                        source_type=SourceType.WEB,
                    )
                )

                successful_crawls += 1
                log_context.info(
                    f"✅ Successfully crawled URL {i}/{len(url_objects)}: {url_obj.source}"
                )

            except Exception as err:
                log_context.debug(f"  ⚠️ Requests failed for {url_obj.source}: {err}")
                docs.append(self._create_error_doc(url_obj, str(err)))

        log_context.info(
            f"📡 Requests crawling: {successful_crawls}/{len(url_objects)} URLs"
        )
        return docs

    def _create_mock_content_from_urls(
        self, url_objects: list[DocObject]
    ) -> list[CrawledObject]:
        """Create mock content from URLs when web crawling fails completely.

        This method generates basic content based on the URL structure to ensure
        the task graph generation can proceed even when web crawling fails.

        Args:
            url_objects (list[DocObject]): List of DocObject instances containing URLs.

        Returns:
            List[CrawledObject]: List of CrawledObject instances with mock content.
        """
        docs: list[CrawledObject] = []

        for url_obj in url_objects:
            url = url_obj.source
            domain = url.split("/")[2] if len(url.split("/")) > 2 else url

            # Extract page type from URL path
            path_parts = url.split("/")
            page_type = "homepage"
            if len(path_parts) > 3:
                page_type = path_parts[3].replace("-", " ").replace("_", " ")

            # Create mock content based on URL structure
            if "company" in url.lower():
                content = f"Company information for {domain}. This page contains details about the company's history, mission, values, and leadership team. The company operates in the robotics and automation industry."
            elif "contact" in url.lower():
                content = f"Contact information for {domain}. This page provides ways to get in touch with the company including phone numbers, email addresses, and office locations."
            elif "privacy" in url.lower():
                content = f"Privacy policy for {domain}. This page outlines how the company collects, uses, and protects user data and personal information."
            elif "terms" in url.lower():
                content = f"Terms and conditions for {domain}. This page contains the legal terms governing the use of the company's services and products."
            elif "resources" in url.lower():
                content = f"Resources and information for {domain}. This page provides additional materials, updates, news, and educational content related to the company's products and services."
            elif "solutions" in url.lower():
                content = f"Solutions and products offered by {domain}. This page showcases the company's robotics and automation solutions including cleaning robots, delivery robots, and multipurpose robots."
            else:
                content = f"Welcome to {domain}. This is the main page providing information about the company's robotics and automation products and services. The company specializes in worker robots, delivery robots, cleaning robots, and multipurpose robots for business applications."

            # Add more context based on the specific URL
            if "clouffee" in url.lower() or "tea" in url.lower():
                content += " This page specifically covers the ClouTea robot milk tea shop operations and tea/coffee making robots."
            elif "headquarters" in url.lower():
                content += " This page provides information about the company's headquarters location and facilities."
            elif "award" in url.lower():
                content += " This page highlights the company's achievements and awards in robotics innovation."
            elif "cleaning" in url.lower():
                content += " This page focuses on cleaning robot solutions including DUST-E SX and DUST-E MX models."
            elif "delivery" in url.lower():
                content += " This page covers delivery robot solutions including Matradee, Matradee X, and Matradee L models."
            elif "production" in url.lower():
                content += (
                    " This page details production and manufacturing robot solutions."
                )

            docs.append(
                CrawledObject(
                    id=url_obj.id,
                    source=url_obj.source,
                    content=content,
                    metadata={
                        "title": f"{page_type.title()} - {domain}",
                        "source": url_obj.source,
                        "mock_content": True,
                    },
                    source_type=SourceType.WEB,
                )
            )

        log_context.info(f"📝 Created mock content for {len(docs)} URLs")
        return docs

    def _create_error_doc(self, url_obj: DocObject, error_msg: str) -> CrawledObject:
        """Create an error document for failed crawls."""
        return CrawledObject(
            id=url_obj.id,
            source=url_obj.source,
            content="",
            metadata={
                "title": url_obj.source,
                "source": url_obj.source,
            },
            is_error=True,
            error_message=error_msg,
            source_type=SourceType.WEB,
        )

    def get_all_urls(self, base_url: str, max_num: int) -> list[str]:
        """Get all URLs from a base URL up to a maximum number.

        This function performs a breadth-first search of URLs starting from a base URL,
        collecting all valid URLs up to the specified maximum number.

        Args:
            base_url (str): The starting URL to crawl from.
            max_num (int): Maximum number of URLs to collect.

        Returns:
            List[str]: List of collected URLs, sorted alphabetically.
        """
        log_context.info(f"🔍 Discovering URLs from {base_url} (max: {max_num})")
        urls_visited = []
        base_url = base_url.split("#")[0].rstrip("/")
        urls_to_visit = [base_url]

        # Add safety limits to prevent infinite loops
        max_iterations = max_num * 3  # Allow some extra iterations for discovery
        iteration_count = 0
        start_time = time.time()
        max_time_seconds = 60  # Maximum 60 seconds for URL discovery

        while urls_to_visit and iteration_count < max_iterations:
            # Check time limit
            if time.time() - start_time > max_time_seconds:
                log_context.warning(
                    f"⏰ URL discovery timed out after {max_time_seconds}s"
                )
                break

            if len(urls_visited) >= max_num:
                break

            current_url = urls_to_visit.pop(0)
            iteration_count += 1

            if current_url not in urls_visited:
                urls_visited.append(current_url)
                try:
                    new_urls = self.get_outsource_urls(current_url, base_url)
                    urls_to_visit.extend(new_urls)
                    urls_to_visit = list(set(urls_to_visit))

                    if new_urls:
                        log_context.info(
                            f"  📎 Found {len(new_urls)} new URLs from {current_url}"
                        )
                except Exception as e:
                    log_context.error(
                        f"  ❌ Error discovering URLs from {current_url}: {e}"
                    )
                    continue

        elapsed_time = time.time() - start_time
        log_context.info(
            f"✅ URL discovery complete: {len(urls_visited)} URLs found in {elapsed_time:.1f}s"
        )
        return sorted(urls_visited[:max_num])

    def get_outsource_urls(self, curr_url: str, base_url: str) -> list[str]:
        """Get outsource URLs from a given URL.

        This function extracts URLs from a webpage that point to external resources.
        It filters and validates the URLs to ensure they are relevant to the base URL.

        Args:
            curr_url (str): The current URL to extract links from.
            base_url (str): The base URL for filtering and validation.

        Returns:
            list[str]: List of valid outsource URLs.
        """
        headers = {
            "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
        }
        new_urls = []
        try:
            response = requests.get(curr_url, headers=headers, timeout=10)
            # Check if the request was successful
            if response.status_code == 200:
                soup = BeautifulSoup(response.text, "html.parser")
                for link in soup.find_all("a"):
                    try:
                        full_url = urljoin(curr_url, link.get("href"))
                        full_url = full_url.split("#")[0].rstrip("/")
                        if self._check_url(full_url, base_url):
                            new_urls.append(full_url)
                    except Exception as err:
                        log_context.error(
                            f"Fail to process sub-url {link.get('href')}: {err}"
                        )
            else:
                log_context.error(
                    f"Failed to retrieve page {curr_url}, status code: {response.status_code}"
                )
        except Exception as err:
            log_context.error(f"Fail to get the page from {curr_url}: {err}")
        return list(set(new_urls))

    def _check_url(self, full_url: str, base_url: str) -> bool:
        """Check if a URL is valid and belongs to the base URL.

        This function validates a URL by checking if it is properly formatted and
        belongs to the specified base URL domain.

        Args:
            full_url (str): The URL to check.
            base_url (str): The base URL for validation.

        Returns:
            bool: True if the URL is valid and belongs to the base URL, False otherwise.
        """
        kw_list = [".pdf", ".jpg", ".png", ".docx", ".xlsx", ".pptx", ".zip", ".jpeg"]
        return (
            full_url.startswith(base_url)
            and full_url
            and not any(kw in full_url for kw in kw_list)
            and full_url != base_url
        )

    def get_candidates_websites(
        self, urls: list[CrawledObject], top_k: int
    ) -> list[CrawledObject]:
        """Get candidate websites based on content relevance.

        This function analyzes the content of crawled URLs and selects the most
        relevant ones based on their content and metadata.

        Args:
            urls (List[CrawledObject]): List of crawled URL objects.
            top_k (int): Number of top candidates to return.

        Returns:
            List[CrawledObject]: List of selected candidate websites.
        """

        nodes = []
        edges = []
        url_to_id_mapping = {}
        for url in urls:
            url_to_id_mapping[url.source] = url.id

        for url in urls:
            if url.is_error:
                continue
            for url_key in url_to_id_mapping:
                if url_key in url.content:
                    edge = [url.id, url_to_id_mapping[url_key]]
                    edges.append(edge)

            node = [url.id, url.to_dict()]
            nodes.append(node)

        self.graph = nx.DiGraph(name="website graph")
        self.graph.add_nodes_from(nodes)
        self.graph.add_edges_from(edges)
        pr = nx.pagerank(self.graph, alpha=0.9)
        # sort the pagerank values in descending order
        sorted_pr = sorted(pr.items(), key=lambda x: x[1], reverse=True)
        log_context.info(f"pagerank results: {sorted_pr}")
        # get the top websites
        top_k_websites = sorted_pr[:top_k]
        urls_candidates = [self.graph.nodes[url_id] for url_id, _ in top_k_websites]
        urls_cleaned = [CrawledObject.from_dict(doc) for doc in urls_candidates if doc]
        return urls_cleaned

    def to_crawled_text(self, text_list: list[str]) -> list[CrawledObject]:
        """Convert a list of text strings to CrawledObject instances.

        This function creates CrawledObject instances from a list of text strings,
        assigning unique IDs and appropriate metadata.

        Args:
            text_list (List[str]): List of text strings to convert.

        Returns:
            List[CrawledObject]: List of CrawledObject instances.
        """
        crawled_local_objs = []
        for text in text_list:
            crawled_obj = CrawledObject(
                id=str(uuid.uuid4()),
                source="text",
                content=text,
                metadata={},
                source_type=SourceType.TEXT,
            )
            crawled_local_objs.append(crawled_obj)
        return crawled_local_objs

    def to_crawled_local_objs(self, file_list: list[str]) -> list[CrawledObject]:
        """Convert a list of local files to CrawledObject instances.

        This function processes local files and creates CrawledObject instances
        for each file, handling different file formats and extracting content.

        Args:
            file_list (List[str]): List of file paths to process.

        Returns:
            List[CrawledObject]: List of CrawledObject instances.
        """
        local_objs = [DocObject(str(uuid.uuid4()), file) for file in file_list]
        crawled_local_objs = [self.crawl_file(local_obj) for local_obj in local_objs]
        return crawled_local_objs

    def crawl_file(self, local_obj: DocObject) -> CrawledObject:
        """Crawl a local file and extract its content.

        This function reads and processes a local file, extracting its content
        and metadata based on the file type. It supports various file formats
        including PDF, Word, Excel, Markdown, and text files.

        Args:
            local_obj (DocObject): The local file object to process.

        Returns:
            CrawledObject: A CrawledObject instance containing the file's content and metadata.
        """
        file_path = Path(local_obj.source)
        file_type = file_path.suffix.lstrip(".")
        file_name = file_path.name

        try:
            if not file_type:
                err_msg = f"No file type detected for file: {str(file_path)}"
                raise FileNotFoundError(err_msg)

            if file_type in ["pdf", "png", "jpg", "jpeg", "pptx", "ppt"] and (
                MISTRAL_API_KEY is not None
                and MISTRAL_API_KEY != "<your-mistral-api-key>"
            ):
                # Call the Mistral API to extract data.
                client = Mistral(api_key=MISTRAL_API_KEY)
                if file_type in ["pdf", "pptx", "ppt"]:
                    with open(file_path, "rb") as file_content:
                        uploaded_doc = client.files.upload(
                            file={
                                "file_name": file_name,
                                "content": file_content,
                            },
                            purpose="ocr",
                        )
                    signed_url = client.files.get_signed_url(file_id=uploaded_doc.id)
                    ocr_response = client.ocr.process(
                        model="mistral-ocr-latest",
                        document={
                            "type": "document_url",
                            "document_url": signed_url.url,
                        },
                    )
                else:
                    # For image files
                    base64_image = encode_image(file_path)
                    ocr_response = client.ocr.process(
                        model="mistral-ocr-latest",
                        document={
                            "type": "image_url",
                            "image_url": f"data:image/{file_type};base64,{base64_image}",
                        },
                    )
                doc_text = ""
                for page in ocr_response.pages:
                    doc_text += page.markdown

                log_context.info("Mistral PDF extractor worked as expected.")
                return CrawledObject(
                    id=local_obj.id,
                    source=local_obj.source,
                    content=doc_text,
                    metadata={"title": file_name, "source": local_obj.source},
                    source_type=SourceType.FILE,
                )
            elif file_type == "html":
                # TODO : Consider replacing this logic with the Unstructured HTML Loader.
                # Would need to be done in crawl_urls too.
                with open(file_path, encoding="utf-8") as f:
                    html = f.read()
                soup = BeautifulSoup(html, "html.parser")

                text_list = []
                for string in soup.strings:
                    if string.find_parent("a"):
                        href = string.find_parent("a").get("href")
                        text = f"{string} {href}"
                        text_list.append(text)
                    elif string.strip():
                        text_list.append(string)
                doc_text = "\n".join(text_list)

                title = file_name
                for title in soup.find_all("title"):
                    title = title.get_text()
                    break

                return CrawledObject(
                    id=local_obj.id,
                    source=local_obj.source,
                    content=doc_text,
                    metadata={"title": title, "source": local_obj.source},
                    source_type=SourceType.FILE,
                )
            elif file_type == "pdf":
                # Since Mistral API key is absent, we default to basic pdf parser
                log_context.info(
                    "MISTRAL_API_KEY env variable not set, hence defaulting to static parsing."
                )
                loader = PyPDFLoader(file_path)
            elif file_type == "doc" or file_type == "docx":
                loader = UnstructuredWordDocumentLoader(file_path, mode="single")
            elif file_type == "xlsx" or file_type == "xls":
                loader = UnstructuredExcelLoader(file_path, mode="single")
            elif file_type == "txt":
                loader = TextLoader(file_path)
            elif file_type == "md":
                loader = UnstructuredMarkdownLoader(file_path)
            elif file_type == "pptx" or file_type == "ppt":
                loader = UnstructuredPowerPointLoader(file_path, mode="single")
            else:
                err_msg = "Unsupported file type. If you are trying to upload a pdf, make sure it is less than 50MB. Images are only supported with the advanced parser."
                raise NotImplementedError(err_msg)

            doc_text = "\n".join(
                [
                    document.to_json()["kwargs"]["page_content"]
                    for document in loader.load()
                ]
            )
            return CrawledObject(
                id=local_obj.id,
                source=local_obj.source,
                content=doc_text,
                metadata={"title": file_name, "source": local_obj.source},
                source_type=SourceType.FILE,
            )

        except Exception as err_msg:
            log_context.info(f"error processing file: {err_msg}")
            return CrawledObject(
                id=local_obj.id,
                source=local_obj.source,
                content=None,
                metadata={"title": file_name},
                source_type=SourceType.FILE,
                is_error=True,
                error_message=str(err_msg),
            )

    @staticmethod
    def save(file_path: str, docs: list[CrawledObject]) -> None:
        """Save a list of CrawledObject instances to a file.

        This function serializes and saves CrawledObject instances to a file
        for later use.

        Args:
            file_path (str): Path where to save the objects.
            docs (List[CrawledObject]): List of CrawledObject instances to save.
        """
        with open(file_path, "wb") as f:
            pickle.dump(docs, f)

    @classmethod
    def chunk(cls, doc_objs: list[CrawledObject]) -> list[CrawledObject]:
        """Split documents into smaller chunks.

        This function splits large documents into smaller, more manageable chunks
        while preserving their metadata and structure.

        Args:
            doc_objs (List[CrawledObject]): List of CrawledObject instances to chunk.

        Returns:
            List[CrawledObject]: List of chunked CrawledObject instances.
        """
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            encoding_name="cl100k_base", chunk_size=200, chunk_overlap=40
        )
        docs = []
        langchain_docs = []
        for doc_obj in doc_objs:
            if doc_obj.is_error or doc_obj.content is None:
                log_context.debug(
                    f"Skip source: {doc_obj.source} because of error or no content"
                )
                continue
            elif doc_obj.is_chunk:
                log_context.debug(
                    f"Skip source: {doc_obj.source} because it has been chunked"
                )
                docs.append(doc_obj)
                continue
            splitted_text = text_splitter.split_text(doc_obj.content)
            for i, txt in enumerate(splitted_text):
                doc = CrawledObject(
                    id=doc_obj.id + "_" + str(i),
                    source=doc_obj.source,
                    content=txt,
                    metadata=doc_obj.metadata,
                    is_chunk=True,
                    source_type=doc_obj.source_type,
                )
                docs.append(doc)
                langchain_docs.append(
                    Document(page_content=txt, metadata={"source": doc_obj.source})
                )
        return langchain_docs
